"""Unaligned Packed Encoding Rules (UPER) codec.

"""

from . import DecodeError
from . import per
from . import restricted_utc_time_to_datetime
from . import restricted_utc_time_from_datetime
from . import restricted_generalized_time_to_datetime
from . import restricted_generalized_time_from_datetime
from .per import to_int
from .per import to_byte_array
from .per import integer_as_number_of_bits
from .per import PermittedAlphabet
from .per import Type
from .per import Boolean
from .per import Real
from .per import Null
from .per import Enumerated
from .per import ObjectIdentifier
from .per import Sequence
from .per import Set
from .per import UTF8String
from .per import GeneralString
from .per import GraphicString
from .per import TeletexString
from .per import UniversalString
from .per import ObjectDescriptor
from .per import Any
from .per import Recursive
from .permitted_alphabet import NUMERIC_STRING
from .permitted_alphabet import PRINTABLE_STRING
from .permitted_alphabet import IA5_STRING
from .permitted_alphabet import BMP_STRING
from .permitted_alphabet import VISIBLE_STRING


class Encoder(per.Encoder):

    def align(self):
        pass


class Decoder(per.Decoder):

    def align(self):
        pass


class KnownMultiplierStringType(per.KnownMultiplierStringType):

    ENCODING = 'ascii'
    PERMITTED_ALPHABET = ''

    def __init__(self,
                 name,
                 minimum=None,
                 maximum=None,
                 has_extension_marker=None,
                 permitted_alphabet=None):
        super(KnownMultiplierStringType, self).__init__(name,
                                                        self.__class__.__name__)
        self.set_size_range(minimum, maximum, has_extension_marker)

        if permitted_alphabet is None:
            permitted_alphabet = self.PERMITTED_ALPHABET

        self.permitted_alphabet = permitted_alphabet
        self.bits_per_character = integer_as_number_of_bits(
            len(permitted_alphabet) - 1)

    def encode(self, data, encoder):
        if self.has_extension_marker:
            encoder.append_bit(0)

        if self.number_of_bits is None:
            return self.encode_unbound(data, encoder)
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(len(data) - self.minimum,
                                                       self.number_of_bits)

        for value in data:
            encoder.append_non_negative_binary_integer(
                self.permitted_alphabet.encode(
                    to_int(value.encode(self.ENCODING))),
                self.bits_per_character)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                raise NotImplementedError(
                    'String size extension is not yet implemented.')

        if self.number_of_bits is None:
            return self.decode_unbound(decoder)
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(self.number_of_bits)

        data = bytearray()

        for _ in range(length):
            value = decoder.read_non_negative_binary_integer(self.bits_per_character)
            value = self.permitted_alphabet.decode(value)
            data += to_byte_array(value, self.bits_per_character)

        return data.decode(self.ENCODING)


class ArrayType(per.ArrayType):

    def encode(self, data, encoder):
        if self.has_extension_marker:
            if self.minimum <= len(data) <= self.maximum:
                encoder.append_bit(0)
            else:
                encoder.append_bit(1)
                encoder.append_length_determinant(len(data))

                for entry in data:
                    self.element_type.encode(entry, encoder)

                return

        if self.number_of_bits is None:
            return self.encode_unbound(data, encoder)
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(len(data) - self.minimum,
                                                       self.number_of_bits)

        for entry in data:
            self.element_type.encode(entry, encoder)

    def decode(self, decoder):
        length = None

        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                length = decoder.read_length_determinant()

        if length is not None:
            pass
        elif self.number_of_bits is None:
            return self.decode_unbound(decoder)
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(
                    self.number_of_bits)

        decoded = []

        for _ in range(length):
            decoded_element = self.element_type.decode(decoder)
            decoded.append(decoded_element)

        return decoded


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name, 'INTEGER')
        self.minimum = None
        self.maximum = None
        self.has_extension_marker = None
        self.number_of_bits = None

    def set_restricted_to_range(self, minimum, maximum, has_extension_marker):
        self.has_extension_marker = has_extension_marker

        if minimum == 'MIN' or maximum == 'MAX':
            return

        self.minimum = minimum
        self.maximum = maximum
        size = self.maximum - self.minimum
        self.number_of_bits = integer_as_number_of_bits(size)

    def encode(self, data, encoder):
        if self.has_extension_marker:
            if self.minimum <= data <= self.maximum:
                encoder.append_bit(0)
            else:
                encoder.append_bit(1)
                encoder.append_unconstrained_whole_number(data)
                return

        if self.number_of_bits is None:
            encoder.append_unconstrained_whole_number(data)
        else:
            encoder.append_non_negative_binary_integer(data - self.minimum,
                                                       self.number_of_bits)

    def decode(self, decoder):
        if self.has_extension_marker:
            if decoder.read_bit():
                return decoder.read_unconstrained_whole_number()

        if self.number_of_bits is None:
            return decoder.read_unconstrained_whole_number()
        else:
            value = decoder.read_non_negative_binary_integer(self.number_of_bits)

            return value + self.minimum

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class BitString(per.BitString):

    def encode(self, data, encoder):
        data, number_of_bits = data

        if self.has_extension_marker:
            if self.minimum <= number_of_bits <= self.maximum:
                encoder.append_bit(0)
            else:
                raise NotImplementedError(
                    'BIT STRING extension is not yet implemented.')

        if self.has_named_bits:
            data, number_of_bits = self.rstrip_zeros(data, number_of_bits)

        if self.number_of_bits is None:
            return self.encode_unbound(data, number_of_bits, encoder)
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(
                number_of_bits - self.minimum,
                self.number_of_bits)

        encoder.append_bits(data, number_of_bits)

    def decode(self, decoder):
        if self.has_extension_marker:
            if decoder.read_bit():
                raise NotImplementedError(
                    'BIT STRING extension is not yet implemented.')

        if self.number_of_bits is None:
            return self.decode_unbound(decoder)
        else:
            number_of_bits = self.minimum

            if self.minimum != self.maximum:
                number_of_bits += decoder.read_non_negative_binary_integer(
                    self.number_of_bits)

        value = decoder.read_bits(number_of_bits)

        return (value, number_of_bits)


class OctetString(per.OctetString):

    def encode(self, data, encoder):
        if self.has_extension_marker:
            if self.minimum <= len(data) <= self.maximum:
                encoder.append_bit(0)
            else:
                encoder.append_bit(1)
                encoder.align()
                encoder.append_length_determinant(len(data))
                encoder.append_bytes(data)

                return

        if self.number_of_bits is None:
            return self.encode_unbound(data, encoder)
        elif self.minimum != self.maximum:
            encoder.append_non_negative_binary_integer(len(data) - self.minimum,
                                                       self.number_of_bits)

        encoder.append_bytes(data)

    def decode(self, decoder):
        if self.has_extension_marker:
            bit = decoder.read_bit()

            if bit:
                length = decoder.read_length_determinant()

                return decoder.read_bytes(length)

        if self.number_of_bits is None:
            return self.decode_unbound(decoder)
        else:
            length = self.minimum

            if self.minimum != self.maximum:
                length += decoder.read_non_negative_binary_integer(
                    self.number_of_bits)

        return decoder.read_bytes(length)


class SequenceOf(ArrayType):

    def __init__(self,
                 name,
                 element_type,
                 minimum,
                 maximum,
                 has_extension_marker):
        super(SequenceOf, self).__init__(name,
                                         element_type,
                                         minimum,
                                         maximum,
                                         has_extension_marker,
                                         'SEQUENCE OF')


class SetOf(ArrayType):

    def __init__(self,
                 name,
                 element_type,
                 minimum,
                 maximum,
                 has_extension_marker):
        super(SetOf, self).__init__(name,
                                    element_type,
                                    minimum,
                                    maximum,
                                    has_extension_marker,
                                    'SET OF')


class Choice(per.Choice):

    def encode_root_index(self, index, encoder):
        encoder.append_non_negative_binary_integer(index, self.root_number_of_bits)

    def decode_root_index(self, decoder):
        return decoder.read_non_negative_binary_integer(self.root_number_of_bits)


class NumericString(KnownMultiplierStringType):

    ALPHABET = bytearray(NUMERIC_STRING.encode('ascii'))
    ENCODE_MAP = {v: i for i, v in enumerate(ALPHABET)}
    DECODE_MAP = {i: v for i, v in enumerate(ALPHABET)}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_MAP,
                                           DECODE_MAP)


class PrintableString(KnownMultiplierStringType):

    ALPHABET = bytearray(PRINTABLE_STRING.encode('ascii'))
    ENCODE_MAP = {v: v for v in ALPHABET}
    DECODE_MAP = {v: v for v in ALPHABET}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_MAP,
                                           DECODE_MAP)


class IA5String(KnownMultiplierStringType):

    ALPHABET = bytearray(IA5_STRING.encode('ascii'))
    ENCODE_DECODE_MAP = {v: v for v in ALPHABET}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_DECODE_MAP,
                                           ENCODE_DECODE_MAP)


class BMPString(KnownMultiplierStringType):

    ENCODING = 'utf-16-be'
    ALPHABET = BMP_STRING
    ENCODE_DECODE_MAP = {ord(v): ord(v) for v in ALPHABET}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_DECODE_MAP,
                                           ENCODE_DECODE_MAP)


class VisibleString(KnownMultiplierStringType):

    ALPHABET = bytearray(VISIBLE_STRING.encode('ascii'))
    ENCODE_DECODE_MAP = {v: v for v in ALPHABET}
    PERMITTED_ALPHABET = PermittedAlphabet(ENCODE_DECODE_MAP,
                                           ENCODE_DECODE_MAP)


class UTCTime(VisibleString):

    def encode(self, data, encoder):
        encoded = restricted_utc_time_from_datetime(data)

        return super(UTCTime, self).encode(encoded, encoder)

    def decode(self, decoder):
        decoded = super(UTCTime, self).decode(decoder)

        return restricted_utc_time_to_datetime(decoded)


class GeneralizedTime(VisibleString):

    def encode(self, data, encoder):
        enceded = restricted_generalized_time_from_datetime(data)

        return super(GeneralizedTime, self).encode(enceded, encoder)

    def decode(self, decoder):
        decoded = super(GeneralizedTime, self).decode(decoder)

        return restricted_generalized_time_to_datetime(decoded)


class Date(per.Date):

    def __init__(self, name):
        super(Date, self).__init__(name)
        immediate = Integer('immediate')
        near_future = Integer('near_future')
        near_past = Integer('near_past')
        reminder = Integer('reminder')
        immediate.set_restricted_to_range(2005, 2020, False)
        near_future.set_restricted_to_range(2021, 2276, False)
        near_past.set_restricted_to_range(1749, 2004, False)
        reminder.set_restricted_to_range('MIN', 1748, False)
        year = Choice('year',
                      [immediate, near_future, near_past, reminder],
                      None)
        month = Integer('month')
        day = Integer('day')
        month.set_restricted_to_range(1, 12, False)
        day.set_restricted_to_range(1, 31, False)
        self._inner = Sequence('DATE-ENCODING',
                               [year, month, day],
                               None)


class TimeOfDay(per.TimeOfDay):

    def __init__(self, name):
        super(TimeOfDay, self).__init__(name)
        hours = Integer('hours')
        minutes = Integer('minutes')
        seconds = Integer('seconds')
        hours.set_restricted_to_range(0, 24, False)
        minutes.set_restricted_to_range(0, 59, False)
        seconds.set_restricted_to_range(0, 60, False)
        self._inner = Sequence('TIME-OF-DAY-ENCODING',
                               [hours, minutes, seconds],
                               None)


class DateTime(per.DateTime):

    def __init__(self, name):
        super(DateTime, self).__init__(name)
        self._inner = Sequence('DATE-TIME-ENCODING',
                               [Date('date'), TimeOfDay('time')],
                               None)


class CompiledType(per.CompiledType):

    def encode(self, data):
        encoder = Encoder()
        self._type.encode(data, encoder)

        return encoder.as_bytearray()

    def decode(self, data):
        decoder = Decoder(bytearray(data))

        return self._type.decode(decoder)


class Compiler(per.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        compiled_type = self.compile_type(type_name,
                                          type_descriptor,
                                          module_name)

        return CompiledType(compiled_type)

    def compile_type(self, name, type_descriptor, module_name):
        module_name = self.get_module_name(type_descriptor, module_name)
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name),
                                  *self.get_size_range(type_descriptor,
                                                       module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name,
                                      sort_by_tag=True))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name),
                             *self.get_size_range(type_descriptor,
                                                  module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(name,
                              *self.compile_members(
                                  type_descriptor['members'],
                                  module_name,
                                  flat_additions=True))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name,
                                  self.get_enum_values(type_descriptor,
                                                       module_name),
                                  self._numeric_enums)
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name,
                                   *self.get_size_range(type_descriptor,
                                                        module_name))
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = NumericString(name,
                                     *self.get_size_range(type_descriptor,
                                                          module_name),
                                     permitted_alphabet=permitted_alphabet)
        elif type_name == 'PrintableString':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = PrintableString(name,
                                       *self.get_size_range(type_descriptor,
                                                            module_name),
                                       permitted_alphabet=permitted_alphabet)
        elif type_name == 'IA5String':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = IA5String(name,
                                 *self.get_size_range(type_descriptor,
                                                      module_name),
                                 permitted_alphabet=permitted_alphabet)
        elif type_name == 'BMPString':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = BMPString(name,
                                 *self.get_size_range(type_descriptor,
                                                      module_name),
                                 permitted_alphabet=permitted_alphabet)
        elif type_name == 'VisibleString':
            permitted_alphabet = self.get_permitted_alphabet(type_descriptor)
            compiled = VisibleString(name,
                                     *self.get_size_range(type_descriptor,
                                                          module_name),
                                     permitted_alphabet=permitted_alphabet)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'DATE':
            compiled = Date(name)
        elif type_name == 'TIME-OF-DAY':
            compiled = TimeOfDay(name)
        elif type_name == 'DATE-TIME':
            compiled = DateTime(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name,
                                 self.get_named_bits(type_descriptor,
                                                     module_name),
                                 *self.get_size_range(type_descriptor,
                                                      module_name))
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        elif type_name == 'EXTERNAL':
            compiled = Sequence(
                name,
                *self.compile_members(self.external_type_descriptor()['members'],
                                      module_name))
        elif type_name == 'ObjectDescriptor':
            compiled = ObjectDescriptor(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        if 'tag' in type_descriptor:
            compiled = self.set_compiled_tag(compiled, type_descriptor)

        if 'restricted-to' in type_descriptor:
            compiled = self.set_compiled_restricted_to(compiled,
                                                       type_descriptor,
                                                       module_name)

        return compiled


def compile_dict(specification, numeric_enums=False):
    return Compiler(specification, numeric_enums).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')



import sys
import json
import unittest
from datetime import date
from datetime import time
from datetime import datetime
from copy import deepcopy
from .utils import Asn1ToolsBaseTest
import asn1tools

sys.path.append('tests/files')

from parameterization import EXPECTED as PARAMETERIZATION


CODECS = ['ber', 'der', 'jer', 'oer', 'per', 'uper', 'xer']
ALL_CODECS = CODECS + ['gser']


def loadb(encoded):
    return json.loads(encoded.decode('utf-8'))


class Asn1ToolsCodecsConsistencyTest(Asn1ToolsBaseTest):

    maxDiff = None

    def encode_decode_all_codecs(self,
                                 type_spec,
                                 values,
                                 numeric_enums=False):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        foos = []

        for codec in CODECS:
            foos.append(asn1tools.compile_string(spec,
                                                 codec,
                                                 numeric_enums=numeric_enums))

        gser = asn1tools.compile_string(spec,
                                        'gser',
                                        numeric_enums=numeric_enums)

        for value in values:
            decoded = value

            for foo in foos:
                encoded = foo.encode('A', decoded)
                decoded = foo.decode('A', encoded)
                self.assertEqual(type(decoded), type(value))
                self.assertEqual(decoded, value)

            gser.encode('A', decoded)

    def encode_decode_codec(self, spec, codec, type_name, decoded, encoded):
        encoded_message = spec.encode(type_name,
                                      decoded,
                                      check_constraints=True)

        if codec == 'jer':
            self.assertEqual(loadb(encoded), loadb(encoded_message))
        else:
            self.assertEqual(encoded_message, encoded)

        if codec == 'gser':
            return

        decoded_message = spec.decode(type_name,
                                      encoded,
                                      check_constraints=True)
        self.assertEqual(decoded_message, decoded)

    def encode_decode_codecs(self,
                             type_spec,
                             decoded,
                             encoded):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        for codec, encoded_message in zip(ALL_CODECS, encoded):
            foo = asn1tools.compile_string(spec, codec)
            self.encode_decode_codec(foo,
                                     codec,
                                     'A',
                                     decoded,
                                     encoded_message)

    def encode_decode_codecs_file(self,
                                  filename,
                                  type_name,
                                  decoded,
                                  encoded):
        for codec, encoded_message in zip(ALL_CODECS, encoded):
            foo = asn1tools.compile_files(filename, codec)
            self.encode_decode_codec(foo,
                                     codec,
                                     type_name,
                                     decoded,
                                     encoded_message)

    def decode_codecs(self, type_spec, decoded, encoded, numeric_enums=False):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= " + type_spec + " "
            + "END"
        )

        for codec, encoded_message in zip(CODECS, encoded):
            foo = asn1tools.compile_string(spec, codec, numeric_enums=numeric_enums)
            decoded_message = foo.decode('A',
                                         encoded_message,
                                         check_constraints=True)
            self.assertEqual(decoded_message, decoded)

    def test_boolean(self):
        self.encode_decode_all_codecs("BOOLEAN", [True, False])

    def test_integer(self):
        self.encode_decode_all_codecs("INTEGER", [1, 123456789, -2, 0])

    def test_integer_versions(self):
        decoded_v1 = 2
        decoded_v2 = 2

        encoded_v2 = [
            b'\x02\x01\x02',
            b'\x02\x01\x02',
            b'2',
            b'\x01\x02',
            b'\x80\x01\x02',
            b'\x80\x81\x00',
            b'<A>2</A>'
        ]

        self.decode_codecs('INTEGER (1, ..., 2)',
                           decoded_v2,
                           encoded_v2)

        self.decode_codecs('INTEGER (1, ...)',
                           decoded_v1,
                           encoded_v2)

    def test_real(self):
        self.encode_decode_all_codecs("REAL", [0.0, 1.0, -1.0])

    def test_null(self):
        self.encode_decode_all_codecs("NULL", [None])

    def test_bit_string(self):
        self.encode_decode_all_codecs("BIT STRING",
                                      [(b'\x58', 5), (b'\x58\x80', 9)])

    def test_octet_string(self):
        self.encode_decode_all_codecs("OCTET STRING", [b'', b'\x12\x34'])

    def test_object_identifier(self):
        self.encode_decode_all_codecs("OBJECT IDENTIFIER", ['1.2.33'])

    def test_enumerated(self):
        self.encode_decode_all_codecs("ENUMERATED { a(0), b(5) }", ['a', 'b'])

    def test_enumerated_numeric(self):
        self.encode_decode_all_codecs("ENUMERATED { a(0), b(5) }",
                                      [0, 5],
                                      numeric_enums=True)

        # Enum with default
        decoded = {"e": 5}

        encoded_messages = [
            b'0\x00',
            b'0\x00',
            b'{}',
            b'\x00',
            b'\x00',
            b'\x00',
            b'<A/>'
        ]

        self.decode_codecs('SEQUENCE { e ENUMERATED { a(0), b(5) } DEFAULT b }', decoded, encoded_messages, numeric_enums=True)

    def test_enumerated_versions(self):
        # Unknown enumeration value 'c' decoded as None.
        decoded_v1 = None
        decoded_v2 = 'c'

        encoded_v2 = [
            b'\x0a\x01\x02',
            b'\x0a\x01\x02',
            b'"c"',
            b'\x02',
            b'\x80',
            b'\x80',
            b'<A><c /></A>'
        ]

        self.decode_codecs('ENUMERATED {a, b, ..., c}',
                           decoded_v2,
                           encoded_v2)

        self.decode_codecs('ENUMERATED {a, b, ...}',
                           decoded_v1,
                           encoded_v2)

    def test_sequence(self):
        self.encode_decode_all_codecs("SEQUENCE { a NULL }", [{'a': None}])

    def test_sequence_of(self):
        self.encode_decode_all_codecs("SEQUENCE OF NULL", [[], [None, None]])

    def test_set(self):
        self.encode_decode_all_codecs("SET { a NULL }", [{'a': None}])

    def test_set_of(self):
        self.encode_decode_all_codecs("SET OF NULL", [[], [None, None]])

    def test_choice(self):
        self.encode_decode_all_codecs("CHOICE { a NULL }", [('a', None)])

    def test_choice_versions(self):
        # Unknown choice 'c' decoded as (None, None).
        decoded_v1 = {
            'a': (None, None),
            'd': True
        }

        decoded_v2 = {
            'a': ('c', None),
            'd': True
        }

        encoded_v2 = [
            b'\x30\x07\xa0\x02\x81\x00\x81\x01\xff',
            b'\x30\x07\xa0\x02\x81\x00\x81\x01\xff',
            b'{"a":{"c":null},"d":true}',
            b'\x81\x00\xff',
            b'\x80\x00\x80',
            b'\x80\x00\x80',
            b'<A><a><c /></a><d><true /></d></A>'
        ]

        self.decode_codecs('SEQUENCE { '
                           '  a CHOICE {b NULL, ..., c NULL}, '
                           '  d BOOLEAN '
                           '}',
                           decoded_v2,
                           encoded_v2)

        self.decode_codecs('SEQUENCE { '
                           '  a CHOICE {b NULL, ...}, '
                           '  d BOOLEAN '
                           '}',
                           decoded_v1,
                           encoded_v2)

    def test_utf8_string(self):
        self.encode_decode_all_codecs("UTF8String", [u'hi'])

    def test_numeric_string(self):
        self.encode_decode_all_codecs("NumericString", [u'123'])

    def test_printable_string(self):
        self.encode_decode_all_codecs("PrintableString", [u'hi'])

    def test_ia5_string(self):
        self.encode_decode_all_codecs("IA5String", [u'hi'])

    def test_visible_string(self):
        self.encode_decode_all_codecs("VisibleString", [u'hi'])

    def test_general_string(self):
        self.encode_decode_all_codecs("GeneralString", [u'hi'])

    def test_bmp_string(self):
        self.encode_decode_all_codecs("BMPString", [u'hi'])

    def test_graphic_string(self):
        self.encode_decode_all_codecs("GraphicString", [u'hi'])

    def test_teletex_string(self):
        self.encode_decode_all_codecs("TeletexString", [u'hi'])

    def test_universal_string(self):
        self.encode_decode_all_codecs("UniversalString", [u'hi'])

    def test_utc_time(self):
        self.encode_decode_all_codecs("UTCTime", [datetime(2020, 3, 12)])

    def test_generalized_time(self):
        self.encode_decode_all_codecs("GeneralizedTime",
                                      [datetime(2021, 3, 12)])

    def test_error_location(self):
        spec = (
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a SEQUENCE { "
            "    b CHOICE { "
            "      c SEQUENCE { "
            "        d INTEGER "
            "      } "
            "    } "
            "  } "
            "}"
            "END"
        )

        for codec in ALL_CODECS:
            foo = asn1tools.compile_string(spec, codec)

            with self.assertRaises(asn1tools.EncodeError) as cm:
                foo.encode('A', {'a': {'b': ('c', {})}})

            self.assertEqual(str(cm.exception),
                             "a: b: c: Sequence member 'd' not found in {}.")

    def test_recursive(self):
        spec = (
            "SEQUENCE { "
            "  a B "
            "} "
            "B ::= CHOICE { "
            "  b A, "
            "  c NULL "
            "} "
        )

        self.encode_decode_all_codecs(spec, [{'a': ('b', {'a': ('c', None)})}])

    def test_with_components(self):
        decoded = {
            'a': 1,
            'b': {
                'c': [
                    (
                        'f', True
                    )
                ],
                'd': 2
            }
        }

        encoded = [
            b'\x30\x0d\x80\x01\x01\xa1\x08\xa0\x03\x81\x01\xff\x81\x01\x02',
            b'\x30\x0d\x80\x01\x01\xa1\x08\xa0\x03\x81\x01\xff\x81\x01\x02',
            b'{"a":1,"b":{"c":[{"f":true}],"d":2}}',
            b'\x01\x01\x01\x81\xff\x01\x02',
            b'\x01\x60\x01\x02',
            b'\x01\x60\x20\x40',
            b'<Foo><a>1</a><b><c><f><true /></f></c><d>2</d></b></Foo>',
            b'foo Foo ::= { a 1, b { c { f : TRUE }, d 2 } }'
        ]

        self.encode_decode_codecs_file('tests/files/with_components.asn',
                                       'Foo',
                                       decoded,
                                       encoded)

    def test_integer_min_max(self):
        decoded = 5

        encoded = [
            b'\x02\x01\x05',
            b'\x02\x01\x05',
            b'5',
            b'\x01\x05',
            b'\x01\x05',
            b'\x01\x05',
            b'<A>5</A>',
            b'a A ::= 5'
        ]

        self.encode_decode_codecs('INTEGER (MIN..MAX)',
                                  decoded,
                                  encoded)

    def test_enumerated_all_except(self):
        decoded = 'c'

        encoded = [
            b'\x0a\x01\x02',
            b'\x0a\x01\x02',
            b'"c"',
            b'\x02',
            b'\x40',
            b'\x40',
            b'<A><c /></A>',
            b'a A ::= c'
        ]

        self.encode_decode_codecs('ENUMERATED {a, b, c, d, e} (ALL EXCEPT b)',
                                  decoded,
                                  encoded)

    def test_constraints_extensions(self):
        decoded = {
            'a': b'\x12\x34',
            'b': b'\x56\x78',
            'c': [True, True, False, True, True, False, True, True, False],
            'd': [True, True, False, True, True, False, True, True, False, True],
            'e': [1, 100, 10000],
            'f': [1, 100, 10000, 1000000]
        }

        encoded = [
            b'\x30\x62\x80\x02\x12\x34\x81\x02\x56\x78\xa2\x1b\x01\x01\xff\x01'
            b'\x01\xff\x01\x01\x00\x01\x01\xff\x01\x01\xff\x01\x01\x00\x01\x01'
            b'\xff\x01\x01\xff\x01\x01\x00\xa3\x1e\x01\x01\xff\x01\x01\xff\x01'
            b'\x01\x00\x01\x01\xff\x01\x01\xff\x01\x01\x00\x01\x01\xff\x01\x01'
            b'\xff\x01\x01\x00\x01\x01\xff\xa4\x0a\x02\x01\x01\x02\x01\x64\x02'
            b'\x02\x27\x10\xa5\x0f\x02\x01\x01\x02\x01\x64\x02\x02\x27\x10\x02'
            b'\x03\x0f\x42\x40',
            b'\x30\x62\x80\x02\x12\x34\x81\x02\x56\x78\xa2\x1b\x01\x01\xff\x01'
            b'\x01\xff\x01\x01\x00\x01\x01\xff\x01\x01\xff\x01\x01\x00\x01\x01'
            b'\xff\x01\x01\xff\x01\x01\x00\xa3\x1e\x01\x01\xff\x01\x01\xff\x01'
            b'\x01\x00\x01\x01\xff\x01\x01\xff\x01\x01\x00\x01\x01\xff\x01\x01'
            b'\xff\x01\x01\x00\x01\x01\xff\xa4\x0a\x02\x01\x01\x02\x01\x64\x02'
            b'\x02\x27\x10\xa5\x0f\x02\x01\x01\x02\x01\x64\x02\x02\x27\x10\x02'
            b'\x03\x0f\x42\x40',
            b'{"a":"1234","b":"5678","c":[true,true,false,true,true,false,true'
            b',true,false],"d":[true,true,false,true,true,false,true,true,fals'
            b'e,true],"e":[1,100,10000],"f":[1,100,10000,1000000]}',
            b'\x02\x12\x34\x02\x56\x78\x01\x09\xff\xff\x00\xff\xff\x00\xff\xff'
            b'\x00\x01\x0a\xff\xff\x00\xff\xff\x00\xff\xff\x00\xff\x01\x03\x01'
            b'\x01\x01\x64\x02\x27\x10\x01\x04\x01\x01\x01\x64\x02\x27\x10\x03'
            b'\x0f\x42\x40',
            b'\x40\x12\x34\x40\x56\x78\x6d\xa0\x0a\xdb\x50\x80\x01\x64\x80\x02'
            b'\x27\x10\x80\x04\x08\x01\x64\x80\x02\x27\x10\x80\x03\x0f\x42\x40',
            b'\x44\x8d\x15\x67\x86\xda\x15\xb6\xa1\x01\x64\x81\x13\x88\x41\x02'
            b'\x02\xc9\x02\x27\x10\x81\x87\xa1\x20\x00',
            b'<A><a>1234</a><b>5678</b><c><true /><true /><false /><true /><tr'
            b'ue /><false /><true /><true /><false /></c><d><true /><true /><f'
            b'alse /><true /><true /><false /><true /><true /><false /><true /'
            b'></d><e><INTEGER>1</INTEGER><INTEGER>100</INTEGER><INTEGER>10000'
            b'</INTEGER></e><f><INTEGER>1</INTEGER><INTEGER>100</INTEGER><INTE'
            b'GER>10000</INTEGER><INTEGER>1000000</INTEGER></f></A>',
            b"a A ::= { a '1234'H, b '5678'H, c { TRUE, TRUE, FALSE, TRUE, TRU"
            b"E, FALSE, TRUE, TRUE, FALSE }, d { TRUE, TRUE, FALSE, TRUE, TRUE"
            b", FALSE, TRUE, TRUE, FALSE, TRUE }, e { 1, 100, 10000 }, f { 1, "
            b"100, 10000, 1000000 } }"
        ]

        self.encode_decode_codecs(
            'SEQUENCE {'
            '  a OCTET STRING (SIZE (1..2, ...)),'
            '  b OCTET STRING (SIZE (1..2), ...),'
            '  c SEQUENCE (SIZE (9..9), ...) OF BOOLEAN,'
            '  d SEQUENCE (SIZE (9..9), ...) OF BOOLEAN,'
            '  e SEQUENCE SIZE (2..3, ...) OF INTEGER (1..5, ...),'
            '  f SEQUENCE (SIZE (2..3, ...)) OF INTEGER (1..5, ...)'
            '}',
            decoded,
            encoded)

    def test_date(self):
        decoded = date(1985, 4, 12)

        encoded = [
            b'\x1f\x1f\x08\x31\x39\x38\x35\x30\x34\x31\x32',
            b'\x1f\x1f\x08\x31\x39\x38\x35\x30\x34\x31\x32',
            b'"1985-04-12"',
            b'\x02\x07\xc1\x04\x0c',
            b'\x80\xec\x35\x80',
            b'\xbb\x0d\x60',
            b'<A>1985-04-12</A>',
            b'a A ::= "1985-04-12"'
        ]

        self.encode_decode_codecs('DATE', decoded, encoded)

    def test_time_of_day(self):
        decoded = time(15, 27, 46)

        encoded = [
            b'\x1f\x20\x06\x31\x35\x32\x37\x34\x36',
            b'\x1f\x20\x06\x31\x35\x32\x37\x34\x36',
            b'"15:27:46"',
            b'\x0f\x1b\x2e',
            b'\x7b\x77\x00',
            b'\x7b\x77\x00',
            b'<A>15:27:46</A>',
            b'a A ::= "15:27:46"'
        ]

        self.encode_decode_codecs('TIME-OF-DAY', decoded, encoded)

    def test_date_time(self):
        decoded = [
            datetime(1985, 4, 12, 15, 27, 46),
            datetime(1582, 2, 3, 4, 5, 6)
        ]

        encoded = [
            [
                b'\x1f\x21\x0e\x31\x39\x38\x35\x30\x34\x31\x32\x31\x35\x32\x37\x34'
                b'\x36',
                b'\x1f\x21\x0e\x31\x39\x38\x35\x30\x34\x31\x32\x31\x35\x32\x37\x34'
                b'\x36',
                b'"1985-04-12T15:27:46"',
                b'\x02\x07\xc1\x04\x0c\x0f\x1b\x2e',
                b'\x80\xec\x35\xbd\xbb\x80',
                b'\xbb\x0d\x6f\x6e\xe0',
                b'<A>1985-04-12T15:27:46</A>',
                b'a A ::= "1985-04-12T15:27:46"'
            ],
            [
                b'\x1f\x21\x0e\x31\x35\x38\x32\x30\x32\x30\x33\x30\x34\x30\x35\x30\x36',
                b'\x1f\x21\x0e\x31\x35\x38\x32\x30\x32\x30\x33\x30\x34\x30\x35\x30\x36',
                b'"1582-02-03T04:05:06"',
                b'\x02\x06\x2e\x02\x03\x04\x05\x06',
                b'\xc0\x02\x06\x2e\x11\x10\x51\x80',
                b'\xc0\x81\x8b\x84\x44\x14\x60',
                b'<A>1582-02-03T04:05:06</A>',
                b'a A ::= "1582-02-03T04:05:06"'
            ]
        ]

        for decoded_message, encoded_message in zip(decoded, encoded):
            self.encode_decode_codecs('DATE-TIME',
                                      decoded_message,
                                      encoded_message)

    def test_named_numbers(self):
        datas = [
            (
                'Constants',
                -1,
                [
                    b'\x02\x01\xff',
                    b'\x02\x01\xff',
                    b'-1',
                    b'\xff',
                    b'\x00',
                    b'\x00',
                    b'<Constants>-1</Constants>'
                ]
            ),
            (
                'Constants',
                2,
                [
                    b'\x02\x01\x02',
                    b'\x02\x01\x02',
                    b'2',
                    b'\x02',
                    b'\xc0',
                    b'\xc0',
                    b'<Constants>2</Constants>'
                ]
            ),
            (
                'A',
                -1,
                [
                    b'\x02\x01\xff',
                    b'\x02\x01\xff',
                    b'-1',
                    b'\xff',
                    b'\x00',
                    b'\x00',
                    b'<A>-1</A>'
                ]
            ),
            (
                'A',
                1,
                [
                    b'\x02\x01\x01',
                    b'\x02\x01\x01',
                    b'1',
                    b'\x01',
                    b'\x80',
                    b'\x80',
                    b'<A>1</A>'
                ]
            ),
            (
                'B',
                2,
                [
                    b'\x02\x01\x02',
                    b'\x02\x01\x02',
                    b'2',
                    b'\x02',
                    b'',
                    b'',
                    b'<B>2</B>'
                ]
            ),
            (
                'C',
                0,
                [
                    b'\x02\x01\x00',
                    b'\x02\x01\x00',
                    b'0',
                    b'\x00',
                    b'\x00',
                    b'\x00',
                    b'<C>0</C>'
                ]
            ),
            (
                'C',
                1,
                [
                    b'\x02\x01\x01',
                    b'\x02\x01\x01',
                    b'1',
                    b'\x01',
                    b'\x80',
                    b'\x80',
                    b'<C>1</C>'
                ]
            )
        ]

        for type_name, decoded, encoded in datas:
            self.encode_decode_codecs_file('tests/files/named_numbers.asn',
                                           type_name,
                                           decoded,
                                           encoded)

    def test_named_numbers_errors(self):
        datas = [
            (
                'Constants',
                -2,
                'Expected an integer between -1 and 2, but got -2.'
            ),
            (
                'Constants',
                3,
                'Expected an integer between -1 and 2, but got 3.'
            ),
            (
                'A',
                -2,
                'Expected an integer between -1 and 1, but got -2.'
            ),
            (
                'A',
                2,
                'Expected an integer between -1 and 1, but got 2.'
            ),
            (
                'B',
                1,
                'Expected an integer between 2 and 2, but got 1.'
            ),
            (
                'C',
                -1,
                'Expected an integer between 0 and 1, but got -1.'
            ),
            (
                'C',
                2,
                'Expected an integer between 0 and 1, but got 2.'
            )
        ]

        for codec in ALL_CODECS:
            foo = asn1tools.compile_files('tests/files/named_numbers.asn',
                                          codec)

            for type_name, decoded, message in datas:
                with self.assertRaises(asn1tools.ConstraintsError) as cm:
                    foo.encode(type_name, decoded, check_constraints=True)

                self.assertEqual(str(cm.exception), message)

    def test_programming_types(self):
        specs = []

        for codec in CODECS:
            specs.append(asn1tools.compile_files(
                'examples/programming_types/programming_types.asn',
                codec))

        decoded = 1

        encoded_messages = [
            b'\x02\x01\x01',
            b'\x02\x01\x01',
            b'1',
            b'\x00\x00\x00\x01',
            b'\xc0\x80\x00\x00\x01',
            b'\x80\x00\x00\x01',
            b'<Int32>1</Int32>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'Int32', decoded, encoded)

    def test_c_source(self):
        specs = []

        for codec in CODECS:
            specs.append(asn1tools.compile_files([
                'tests/files/c_source/c_source.asn',
                'examples/programming_types/programming_types.asn'
            ], codec))

        # Type A.
        decoded = {
            'a': -1,
            'b': -2,
            'c': -3,
            'd': -4,
            'e': 1,
            'f': 2,
            'g': 3,
            'h': 4,
            'i': True,
            'j': 11 * b'\x05'
        }

        encoded_messages = [
            b'\x30\x28\x80\x01\xff\x81\x01\xfe\x82\x01\xfd\x83\x01\xfc\x84'
            b'\x01\x01\x85\x01\x02\x86\x01\x03\x87\x01\x04\x88\x01\xff\x89'
            b'\x0b\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'\x30\x28\x80\x01\xff\x81\x01\xfe\x82\x01\xfd\x83\x01\xfc\x84'
            b'\x01\x01\x85\x01\x02\x86\x01\x03\x87\x01\x04\x88\x01\xff\x89'
            b'\x0b\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'{"c":-3,"f":2,"d":-4,"a":-1,"i":true,"e":1,"j":"050505050505'
            b'0505050505","h":4,"b":-2,"g":3}',
            b'\xff\xff\xfe\xff\xff\xff\xfd\xff\xff\xff\xff\xff\xff\xff\xfc'
            b'\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x04'
            b'\xff\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'\x7f\x7f\xfe\xc0\x7f\xff\xff\xfd\xe0\x7f\xff\xff\xff\xff\xff'
            b'\xff\xfc\x01\x00\x02\x00\x03\x00\x04\x80\x05\x05\x05\x05\x05'
            b'\x05\x05\x05\x05\x05\x05',
            b'\x7f\x7f\xfe\x7f\xff\xff\xfd\x7f\xff\xff\xff\xff\xff\xff\xfc'
            b'\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00\x04'
            b'\x82\x82\x82\x82\x82\x82\x82\x82\x82\x82\x82\x80',
            b'<A><a>-1</a><b>-2</b><c>-3</c><d>-4</d><e>1</e><f>2</f><g>3<'
            b'/g><h>4</h><i><true /></i><j>0505050505050505050505</j></A>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'A', decoded, encoded)

        # Type B, choice a.
        decoded = ('a', -10)

        encoded_messages = [
            b'\x80\x01\xf6',
            b'\x80\x01\xf6',
            b'{"a": -10}',
            b'\x80\xf6',
            b'\x00\x76',
            b'\x1d\x80',
            b'<B><a>-10</a></B>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'B', decoded, encoded)

        # Type B, choice b.
        decoded = (
            'b',
            {
                'a': -1,
                'b': -2,
                'c': -3,
                'd': -4,
                'e': 1,
                'f': 2,
                'g': 3,
                'h': 4,
                'i': True,
                'j': 11 * b'\x05'
            }
        )

        encoded_messages = [
            b'\xa1\x28\x80\x01\xff\x81\x01\xfe\x82\x01\xfd\x83\x01\xfc\x84'
            b'\x01\x01\x85\x01\x02\x86\x01\x03\x87\x01\x04\x88\x01\xff\x89'
            b'\x0b\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'\xa1\x28\x80\x01\xff\x81\x01\xfe\x82\x01\xfd\x83\x01\xfc\x84'
            b'\x01\x01\x85\x01\x02\x86\x01\x03\x87\x01\x04\x88\x01\xff\x89'
            b'\x0b\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'{"b":{"h":4,"f":2,"i":true,"d":-4,"a":-1,"j":"05050505050505'
            b'05050505","e":1,"g":3,"c":-3,"b":-2}}',
            b'\x81\xff\xff\xfe\xff\xff\xff\xfd\xff\xff\xff\xff\xff\xff\xff'
            b'\xfc\x01\x00\x02\x00\x00\x00\x03\x00\x00\x00\x00\x00\x00\x00'
            b'\x04\xff\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05\x05',
            b'\x40\x7f\x7f\xfe\xc0\x7f\xff\xff\xfd\xe0\x7f\xff\xff\xff\xff'
            b'\xff\xff\xfc\x01\x00\x02\x00\x03\x00\x04\x80\x05\x05\x05\x05'
            b'\x05\x05\x05\x05\x05\x05\x05',
            b'\x5f\xdf\xff\x9f\xff\xff\xff\x5f\xff\xff\xff\xff\xff\xff\xff'
            b'\x00\x40\x00\x80\x00\x00\x00\xc0\x00\x00\x00\x00\x00\x00\x01'
            b'\x20\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0\xa0',
            b'<B><b><a>-1</a><b>-2</b><c>-3</c><d>-4</d><e>1</e><f>2</f><g'
            b'>3</g><h>4</h><i><true /></i><j>0505050505050505050505</j></'
            b'b></B>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'B', decoded, encoded)

        # Type C - empty.
        decoded = []

        encoded_messages = [
            b'\x30\x00',
            b'\x30\x00',
            b'[]',
            b'\x01\x00',
            b'\x00',
            b'\x00',
            b'<C />'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'C', decoded, encoded)

        # Type C - 2 elements.
        decoded = [('a', -11), ('a', 13)]

        encoded_messages = [
            b'\x30\x06\x80\x01\xf5\x80\x01\x0d',
            b'\x30\x06\x80\x01\xf5\x80\x01\x0d',
            b'[{"a": -11}, {"a": 13}]',
            b'\x01\x02\x80\xf5\x80\x0d',
            b'\x80\x75\x00\x8d',
            b'\x87\x52\x34',
            b'<C><a>-11</a><a>13</a></C>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'C', decoded, encoded)

        # Type D.
        decoded = [
            {
                'a': {
                    'b': ('c', 0),
                    'e': [None, None, None],
                    'f': None
                },
                'g': {
                    'h': 'j',
                    'l': b'\x54\x55'
                },
                'm': {
                    'n': False,
                    'o': 2,
                    'p': {
                        'q': 5 * b'\x03',
                        'r': True
                    },
                    's': True
                }
            }
        ]

        encoded_messages = [
            b'\x30\x30\x30\x2e\xa0\x0f\xa0\x03\x80\x01\x00\xa1\x06\x05\x00\x05'
            b'\x00\x05\x00\x82\x00\xa1\x04\x81\x02\x54\x55\xa2\x15\x80\x01\x00'
            b'\x81\x01\x02\xa2\x0a\x80\x05\x03\x03\x03\x03\x03\x81\x01\xff\x83'
            b'\x01\xff',
            b'\x30\x30\x30\x2e\xa0\x0f\xa0\x03\x80\x01\x00\xa1\x06\x05\x00\x05'
            b'\x00\x05\x00\x82\x00\xa1\x04\x81\x02\x54\x55\xa2\x15\x80\x01\x00'
            b'\x81\x01\x02\xa2\x0a\x80\x05\x03\x03\x03\x03\x03\x81\x01\xff\x83'
            b'\x01\xff',
            b'[{"m":{"p":{"q":"0303030303","r":true},"s":true,"o":2,"n":false}'
            b',"g":{"l":"5455","h":"j"},"a":{"b":{"c":0},"e":[null,null,null],'
            b'"f":null}}]',
            b'\x01\x01\x80\x00\x01\x03\x00\x02\x54\x55\xf0\x00\x02\x80\x03\x03'
            b'\x03\x03\x03\xff\xff',
            b'\x00\x80\x54\x55\xf4\x80\x03\x03\x03\x03\x03\xc0',
            b'\x00\xaa\x2a\xfa\x40\xc0\xc0\xc0\xc0\xf0',
            b'<D><SEQUENCE><a><b><c>0</c></b><e><NULL /><NULL /><NULL /></e><f'
            b' /></a><g><h><j /></h><l>5455</l></g><m><n><false /></n><o>2</o>'
            b'<p><q>0303030303</q><r><true /></r></p><s><true /></s></m></SEQU'
            b'ENCE></D>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'D', decoded, encoded)

        # Type D - some missing.
        decoded = [
            {
                'a': {
                    'b': ('d', False),
                    'e': [None, None, None],
                    'f': None
                },
                'g': {
                    'h': 'k',
                    'l': b'\x54'
                },
                'm': {
                    'o': 3,
                    'p': {
                        'q': 5 * b'\x03'
                    },
                    's': False
                }
            }
        ]

        encoded_messages = [
            b'\x30\x27\x30\x25\xa0\x0f\xa0\x03\x81\x01\x00\xa1\x06\x05\x00\x05'
            b'\x00\x05\x00\x82\x00\xa1\x07\x80\x02\x02\x00\x81\x01\x54\xa2\x09'
            b'\xa2\x07\x80\x05\x03\x03\x03\x03\x03',
            b'\x30\x27\x30\x25\xa0\x0f\xa0\x03\x81\x01\x00\xa1\x06\x05\x00\x05'
            b'\x00\x05\x00\x82\x00\xa1\x07\x80\x02\x02\x00\x81\x01\x54\xa2\x09'
            b'\xa2\x07\x80\x05\x03\x03\x03\x03\x03',
            b'[{"a":{"b":{"d":false},"e":[null,null,null],"f":null},"g":{"h":"'
            b'k","l":"54"},"m":{"o":3,"p":{"q":"0303030303"},"s":false}}]',
            b'\x01\x01\x81\x00\x01\x03\x80\x82\x02\x00\x01\x54\x20\x00\x03\x03'
            b'\x03\x03\x03',
            b'\x09\x80\x54\x20\x03\x03\x03\x03\x03',
            b'\x09\x8a\x84\x03\x03\x03\x03\x03',
            b'<D><SEQUENCE><a><b><d><false /></d></b><e><NULL /><NULL /><NULL '
            b'/></e><f /></a><g><h><k /></h><l>54</l></g><m><o>3</o><p><q>0303'
            b'030303</q></p><s><false /></s></m></SEQUENCE></D>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'D', decoded, encoded)

        # Type E.
        decoded = {'a': ('b', ('c', True))}

        encoded_messages = [
            b'\x30\x07\xa0\x05\xa0\x03\x80\x01\xff',
            b'\x30\x07\xa0\x05\xa0\x03\x80\x01\xff',
            b'{"a": {"b": {"c": true}}}',
            b'\x80\x80\xff',
            b'\x80',
            b'\x80',
            b'<E><a><b><c><true /></c></b></a></E>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'E', decoded, encoded)

        # Type F.
        decoded = [[False], [True]]

        encoded_messages = [
            b'0\n0\x03\x01\x01\x000\x03\x01\x01\xff',
            b'0\n0\x03\x01\x01\x000\x03\x01\x01\xff',
            b'[[false], [true]]',
            b'\x01\x02\x01\x01\x00\x01\x01\xff',
            b'\xa0',
            b'\xa0',
            b'<F><SEQUENCE_OF><false /></SEQUENCE_OF><SEQUENCE_OF><true /></SE'
            b'QUENCE_OF></F>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'F', decoded, encoded)

        # Type G.
        decoded = {'a': True, 'i': True}

        encoded_messages = [
            b'\x30\x06\x80\x01\xff\x88\x01\xff',
            b'\x30\x06\x80\x01\xff\x88\x01\xff',
            b'{"a": true, "i": true}',
            b'\x80\x80\xff\xff',
            b'\x80\xe0',
            b'\x80\xe0',
            b'<G><a><true /></a><i><true /></i></G>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'G', decoded, encoded)

        # Type H.
        decoded = None

        encoded_messages = [
            b'\x05\x00',
            b'\x05\x00',
            b'null',
            b'',
            b'',
            b'',
            b'<H />'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'H', decoded, encoded)

        # Type I.
        decoded = (
            b'\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03'
            b'\x04\x01\x02\x03\x04\x01\x02\x03\x04'
        )

        encoded_messages = [
            b'\x04\x18\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01'
            b'\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04',
            b'\x04\x18\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01'
            b'\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04',
            b'"010203040102030401020304010203040102030401020304"',
            b'\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03'
            b'\x04\x01\x02\x03\x04\x01\x02\x03\x04',
            b'\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03'
            b'\x04\x01\x02\x03\x04\x01\x02\x03\x04',
            b'\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03'
            b'\x04\x01\x02\x03\x04\x01\x02\x03\x04',
            b'<I>010203040102030401020304010203040102030401020304</I>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'I', decoded, encoded)

        # Type J.
        decoded = (
            b'\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03'
            b'\x04\x01\x02\x03\x04\x01\x02'
        )

        encoded_messages = [
            b'\x04\x16\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01'
            b'\x02\x03\x04\x01\x02\x03\x04\x01\x02',
            b'\x04\x16\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01'
            b'\x02\x03\x04\x01\x02\x03\x04\x01\x02',
            b'"01020304010203040102030401020304010203040102"',
            b'\x16\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02'
            b'\x03\x04\x01\x02\x03\x04\x01\x02',
            b'\x00\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02\x03\x04\x01\x02'
            b'\x03\x04\x01\x02\x03\x04\x01\x02',
            b'\x00\x81\x01\x82\x00\x81\x01\x82\x00\x81\x01\x82\x00\x81\x01'
            b'\x82\x00\x81\x01\x82\x00\x81\x00',
            b'<J>01020304010203040102030401020304010203040102</J>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'J', decoded, encoded)

        # Type K.
        decoded = 'a'

        encoded_messages = [
            b'\x0a\x01\x00',
            b'\x0a\x01\x00',
            b'"a"',
            b'\x00',
            b'',
            b'',
            b'<K><a /></K>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'K', decoded, encoded)

        # Type L.
        decoded = 260 * b'\xa5'

        encoded_messages = [
            b'\x04\x82\x01\x04' + 260 * b'\xa5',
            b'\x04\x82\x01\x04' + 260 * b'\xa5',
            b'"' + 260 * b'A5' + b'"',
            b'\x82\x01\x04' + 260 * b'\xa5',
            b'\x01\x04' + 260 * b'\xa5',
            b'\x82\x52' + 259 * b'\xd2' + b'\x80',
            b'<L>' + 260 * b'A5' + b'</L>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'L', decoded, encoded)

        # Type O.
        decoded = 260 * [True]

        encoded_messages = [
            b'\x30\x82\x03\x0c' + 260 * b'\x01\x01\xff',
            b'\x30\x82\x03\x0c' + 260 * b'\x01\x01\xff',
            b'[' + 259 * b'true,' + b'true]',
            b'\x02\x01\x04' + 260 * b'\xff',
            b'\x01\x03' + 32 * b'\xff' + b'\xf0',
            b'\x81' + 32 * b'\xff' + b'\xf8',
            b'<O>' + 260 * b'<true />' + b'</O>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'O', decoded, encoded)

        # Type Q.
        decoded = ('c256', True)

        encoded_messages = [
            b'\x9f\x81\x7f\x01\xff',
            b'\x9f\x81\x7f\x01\xff',
            b'{"c256": true}',
            b'\xbf\x81\x7f\xff',
            b'\x00\xff\x80',
            b'\x7f\xc0',
            b'<Q><c256><true /></c256></Q>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'Q', decoded, encoded)

        # Type Q.
        decoded = ('c257', True)

        encoded_messages = [
            b'\x9f\x82\x00\x01\xff',
            b'\x9f\x82\x00\x01\xff',
            b'{"c257": true}',
            b'\xbf\x82\x00\xff',
            b'\x01\x00\x80',
            b'\x80\x40',
            b'<Q><c257><true /></c257></Q>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'Q', decoded, encoded)

        # Type R.
        decoded = -1

        encoded_messages = [
            b'\x02\x01\xff',
            b'\x02\x01\xff',
            b'-1',
            b'\xff',
            b'\x00',
            b'\x00',
            b'<R>-1</R>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'R', decoded, encoded)

        decoded = 0

        encoded_messages = [
            b'\x02\x01\x00',
            b'\x02\x01\x00',
            b'0',
            b'\x00',
            b'\x80',
            b'\x80',
            b'<R>0</R>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'R', decoded, encoded)

        # Type S.
        decoded = -2

        encoded_messages = [
            b'\x02\x01\xfe',
            b'\x02\x01\xfe',
            b'-2',
            b'\xfe',
            b'\x00',
            b'\x00',
            b'<S>-2</S>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'S', decoded, encoded)

        decoded = 1

        encoded_messages = [
            b'\x02\x01\x01',
            b'\x02\x01\x01',
            b'1',
            b'\x01',
            b'\xc0',
            b'\xc0',
            b'<S>1</S>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'S', decoded, encoded)

        # Type T.
        decoded = -1

        encoded_messages = [
            b'\x02\x01\xff',
            b'\x02\x01\xff',
            b'-1',
            b'\xff',
            b'\x00',
            b'\x00',
            b'<T>-1</T>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'T', decoded, encoded)

        decoded = 2

        encoded_messages = [
            b'\x02\x01\x02',
            b'\x02\x01\x02',
            b'2',
            b'\x02',
            b'\xc0',
            b'\xc0',
            b'<T>2</T>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'T', decoded, encoded)

        # Type U.
        decoded = -64

        encoded_messages = [
            b'\x02\x01\xc0',
            b'\x02\x01\xc0',
            b'-64',
            b'\xc0',
            b'\x00',
            b'\x00',
            b'<U>-64</U>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'U', decoded, encoded)

        # Type V.
        decoded = -128

        encoded_messages = [
            b'\x02\x01\x80',
            b'\x02\x01\x80',
            b'-128',
            b'\x80',
            b'\x00',
            b'\x00',
            b'<V>-128</V>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'V', decoded, encoded)

        # Type W.
        decoded = -1

        encoded_messages = [
            b'\x02\x01\xff',
            b'\x02\x01\xff',
            b'-1',
            b'\xff\xff',
            b'\x00\x00',
            b'\x00\x00',
            b'<W>-1</W>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'W', decoded, encoded)

        decoded = 510

        encoded_messages = [
            b'\x02\x02\x01\xfe',
            b'\x02\x02\x01\xfe',
            b'510',
            b'\x01\xfe',
            b'\x01\xff',
            b'\xff\x80',
            b'<W>510</W>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'W', decoded, encoded)

        # Type X.
        decoded = -2

        encoded_messages = [
            b'\x02\x01\xfe',
            b'\x02\x01\xfe',
            b'-2',
            b'\xff\xfe',
            b'\x00\x00',
            b'\x00\x00',
            b'<X>-2</X>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'X', decoded, encoded)

        decoded = 510

        encoded_messages = [
            b'\x02\x02\x01\xfe',
            b'\x02\x02\x01\xfe',
            b'510',
            b'\x01\xfe',
            b'\x02\x00',
            b'\x80\x00',
            b'<X>510</X>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'X', decoded, encoded)

        # Type Y.
        decoded = 10000

        encoded_messages = [
            b'\x02\x02\x27\x10',
            b'\x02\x02\x27\x10',
            b'10000',
            b'\x27\x10',
            b'\x00\x00',
            b'\x00\x00',
            b'<Y>10000</Y>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'Y', decoded, encoded)

        decoded = 10512

        encoded_messages = [
            b'\x02\x02\x29\x10',
            b'\x02\x02\x29\x10',
            b'10512',
            b'\x29\x10',
            b'\x02\x00',
            b'\x80\x00',
            b'<Y>10512</Y>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'Y', decoded, encoded)

        # AB.
        decoded = {
            'a': 0,
            'b': 10300
        }

        encoded_messages = [
            b'\x30\x0b\xa0\x03\x02\x01\x00\xa1\x04\x02\x02\x28\x3c',
            b'\x30\x0b\xa0\x03\x02\x01\x00\xa1\x04\x02\x02\x28\x3c',
            b'{"a": 0, "b": 10300}',
            b'\x00\x28\x3c',
            b'\x80\x01,',
            b'\xa5\x80',
            b'<AB><a>0</a><b>10300</b></AB>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AB', decoded, encoded)

        # AE
        decoded = {
            'a': False,
            'b': True,
            'c': False
        }

        encoded_messages = [
            b'\x30\x06\x80\x01\x00\x82\x01\x00',
            b'\x30\x06\x80\x01\x00\x82\x01\x00',
            b'{"a": false, "b": true, "c": false}',
            b'\x40\x00\x00',
            b'\x40',
            b'\x40',
            b'<AE><a><false /></a><b><true /></b><c><false /></c></AE>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AE', decoded, encoded)

        # AF
        decoded = {
            'a': True,
            'b': {
                'c': True,
                'd': 17,
                'e': 'g'
            },
            'e': 18,
            'f': 19,
            'g': 20,
            'h': 21,
            'i': 22,
            'j': 23,
            'k': 24,
            'l': 25,
        }

        encoded_messages = [
            b'\x30\x23\x80\x01\xff\xa1\x09\x80\x01\xff\x81\x01\x11\x82\x01\x01'
            b'\x82\x01\x12\x83\x01\x13\x84\x01\x14\x85\x01\x15\x86\x01\x16\x87'
            b'\x01\x17\x89\x01\x19',
            b'\x30\x23\x80\x01\xff\xa1\x09\x80\x01\xff\x81\x01\x11\x82\x01\x01'
            b'\x82\x01\x12\x83\x01\x13\x84\x01\x14\x85\x01\x15\x86\x01\x16\x87'
            b'\x01\x17\x89\x01\x19',
            b'{"a": true, "b": {"c":true, "d":17, "e":"g"}, "e":18, "f":19, "g":20,'
            b'"h":21, "i":22, "j":23, "k":24, "l":25}',
            b'\x80\xff\x03\x07\xff\x80\x09\x80\xff\x02\x06\xc0\x01\x11\x01\x01'
            b'\x01\x12\x01\x13\x01\x14\x01\x15\x01\x16\x01\x17\x01\x18\x01\x19',
            b'\xc4\x7f\xc0\x06\xc0\xe0\x01\x11\x01\x40\x01\x12\x01\x13\x01\x14'
            b'\x01\x15\x01\x16\x01\x17\x01\x18\x01\x19',
            b'\xc4\x7f\xc1\xb0\x38\x08\x88\x0a\x00\x00\x44\x80\x44\xc0\x45\x00'
            b'\x45\x40\x45\x80\x45\xc0\x46\x00\x46\x40',
            b'<AF><a><true /></a><b><c><true /></c><d>17</d><e><g /></e></b><e>18'
            b'</e><f>19</f><g>20</g><h>21</h><i>22</i><j>23</j><k>24</k><l>25'
            b'</l></AF>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AF', decoded, encoded)

        # Type AG
        decoded = {
            'a': True,
            'b': b'\x84\x55',
            'c': [True, False, True, False],
            'd': 'f',
            'h': None,
            'i': 1.0,
            'j': ('k', 60693),
            'm': b'\xf0\xf1\xf2\xf3\xf4'
        }

        encoded_messages = [
            b'\x30\x2e\x80\x01\xff\x81\x02\x84\x55\xa2\x0c\x01\x01\xff\x01\x01'
            b'\x00\x01\x01\xff\x01\x01\x00\x83\x02\x01\x00\x84\x00\x85\x03\x80'
            b'\x00\x01\xa6\x05\x80\x03\x00\xed\x15\x87\x05\xf0\xf1\xf2\xf3\xf4',
            b'\x30\x2e\x80\x01\xff\x81\x02\x84\x55\xa2\x0c\x01\x01\xff\x01\x01'
            b'\x00\x01\x01\xff\x01\x01\x00\x83\x02\x01\x00\x84\x00\x85\x03\x80'
            b'\x00\x01\xa6\x05\x80\x03\x00\xed\x15\x87\x05\xf0\xf1\xf2\xf3\xf4',
            b'{"a": true, "b": "8455", "c": [true, false, true, false], "d": "f",'
            b'"h": null, "i": 1.0, "j": {"k": 60693}, "m": "F0F1F2F3F4"}',
            b'\x80\xff\x02\x01\xfe\x03\x02\x84\x55\x06\x01\x04\xff\x00\xff\x00'
            b'\x03\x82\x01\x00\x00\x04\x3f\x80\x00\x00\x03\x80\xed\x15\x05\xf0'
            b'\xf1\xf2\xf3\xf4',
            b'\xc3\x7f\x03\x20\x84\x55\x01\x3a\x01\x40\x00\x04\x03\x80\x00\x01'
            b'\x03\x00\xed\x15\x05\xf0\xf1\xf2\xf3\xf4',
            b'\xc3\x7f\x03\x28\x45\x50\x01\x3a\x01\x40\x00\x04\x03\x80\x00\x01'
            b'\x03\x76\x8a\x80\x05\xf0\xf1\xf2\xf3\xf4',
            b'<AG><a><true /></a><b>8455</b><c><true /><false /><true /><false />'
            b'</c><d><f /></d><h /><i>1.0E0</i><j><k>60693</k></j><m>F0F1F2F3F4</'
            b'm></AG>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AG', decoded, encoded)

        # Type AL.
        decoded = -129

        encoded_messages = [
            b'\x02\x02\xff\x7f',
            b'\x02\x02\xff\x7f',
            b'-129',
            b'\xff\x7f',
            b'\x00\x00',
            b'\x00\x00',
            b'<AL>-129</AL>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AL', decoded, encoded)

        decoded = 127

        encoded_messages = [
            b'\x02\x01\x7f',
            b'\x02\x01\x7f',
            b'127',
            b'\x00\x7f',
            b'\x01\x00',
            b'\x80\x00',
            b'<AL>127</AL>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AL', decoded, encoded)

        # Type AM.
        decoded = -2

        encoded_messages = [
            b'\x02\x01\xfe',
            b'\x02\x01\xfe',
            b'-2',
            b'\xff\xfe',
            b'\x00',
            b'\x00',
            b'<AM>-2</AM>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AM', decoded, encoded)

        decoded = 128

        encoded_messages = [
            b'\x02\x02\x00\x80',
            b'\x02\x02\x00\x80',
            b'128',
            b'\x00\x80',
            b'\x82',
            b'\x82',
            b'<AM>128</AM>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AM', decoded, encoded)

        # Type AQ.
        decoded = 1234

        encoded_messages = [
            b'\x02\x02\x04\xd2',
            b'\x02\x02\x04\xd2',
            b'1234',
            b'\x00\x00\x04\xd2',
            b'\x40\x04\xd2',
            b'\x00\x04\xd2',
            b'<AQ>1234</AQ>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AQ', decoded, encoded)

        decoded = 100001

        encoded_messages = [
            b'\x02\x03\x01\x86\xa1',
            b'\x02\x03\x01\x86\xa1',
            b'100001',
            b'\x00\x01\x86\xa1',
            b'\x80\x01\x86\xa1',
            b'\x01\x86\xa1',
            b'<AQ>100001</AQ>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'AQ', decoded, encoded)

    def test_parameterization(self):
        specs = []

        for codec in CODECS:
            specs.append(asn1tools.compile_dict(deepcopy(PARAMETERIZATION),
                                                codec))

        # A-Boolean.
        decoded = {'a': True}

        encoded_messages = [
            b'\x30\x05\xa0\x03\x01\x01\xff',
            b'\x30\x05\xa0\x03\x01\x01\xff',
            b'{"a": true}',
            b'\xff',
            b'\x80',
            b'\x80',
            b'<A-Boolean><a><true /></a></A-Boolean>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'A-Boolean', decoded, encoded)

        # A-Integer.
        decoded = {'a': 555}

        encoded_messages = [
            b'\x30\x06\xa0\x04\x02\x02\x02\x2b',
            b'\x30\x06\xa0\x04\x02\x02\x02\x2b',
            b'{"a": 555}',
            b'\x02\x02\x2b',
            b'\x02\x02\x2b',
            b'\x02\x02\x2b',
            b'<A-Integer><a>555</a></A-Integer>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'A-Integer', decoded, encoded)

        # B-BooleanInteger.
        decoded = {
            'a': True,
            'b': -40000
        }

        encoded_messages = [
            b'\x30\x0c\xa0\x03\x01\x01\xff\xa1\x05\x02\x03\xff\x63\xc0',
            b'\x30\x0c\xa0\x03\x01\x01\xff\xa1\x05\x02\x03\xff\x63\xc0',
            b'{"a": true, "b": -40000}',
            b'\x80\xff\x03\xff\x63\xc0',
            b'\xc0\x03\xff\x63\xc0',
            b'\xc0\xff\xd8\xf0\x00',
            b'<B-BooleanInteger><a><true /></a><b>-40000</b></B-BooleanInteger>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec,
                                     codec,
                                     'B-BooleanInteger',
                                     decoded,
                                     encoded)

        # D.
        decoded = {
            'a': ('a', {'a': True}),
            'b': ('c', {'a': {'a': None, 'b': 5}})
        }

        encoded_messages = [
            b'\x30\x1c\xa0\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1\x0f\xa0'
            b'\x0d\xa0\x0b\x30\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'\x30\x1c\xa0\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1\x0f\xa0'
            b'\x0d\xa0\x0b\x30\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'{"a": {"a": {"a": true}}, "b": {"c": {"a": {"a": null, "b": 5}}}'
            b'}',
            b'\x80\x00\xff\x80\x80\x01\x05',
            b'\x28\x01\x05',
            b'\x28\x08\x28',
            b'<D><a><a><a><true /></a></a></a><b><c><a><a /><b>5</b></a></c></'
            b'b></D>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'D', decoded, encoded)

        decoded = {
            'a': ('a', {'a': True}),
            'b': ('d', {'a': None, 'b': 5})
        }

        encoded_messages = [
            b'\x30\x18\xa0\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1\x0b\xa1'
            b'\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'\x30\x18\xa0\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1\x0b\xa1'
            b'\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'{"a": {"a": {"a": true}}, "b": {"d": {"a": null, "b": 5}}}',
            b'\x80\x00\xff\x81\x80\x01\x05',
            b'\x38\x01\x05',
            b'\x38\x08\x28',
            b'<D><a><a><a><true /></a></a></a><b><d><a /><b>5</b></d></b></D>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'D', decoded, encoded)

        decoded = {
            'a': ('b', {'a': {'a': True}}),
            'b': ('d', {'a': None, 'b': 5})
        }

        encoded_messages = [
            b'\x30\x1a\xa0\x0b\xa1\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1'
            b'\x0b\xa1\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'\x30\x1a\xa0\x0b\xa1\x09\xa0\x07\x30\x05\xa0\x03\x01\x01\xff\xa1'
            b'\x0b\xa1\x09\xa0\x02\x05\x00\xa1\x03\x02\x01\x05',
            b'{"a": {"b": {"a": {"a": true}}}, "b": {"d": {"a": null, "b": 5}}}',
            b'\x81\x00\x00\xff\x81\x80\x01\x05',
            b'\x9c\x01\x05',
            b'\x9c\x04\x14',
            b'<D><a><b><a><a><true /></a></a></b></a><b><d><a /><b>5</b></d></b'
            b'></D>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'D', decoded, encoded)

        # F.
        decoded = [('b', {'c': [{'a': 2, 'b': True}]})]

        encoded_messages = [
            b'\x30\x10\xa1\x0e\xa0\x0c\x30\x0a\xa0\x03\x02\x01\x02\xa1\x03\x01'
            b'\x01\xff',
            b'\x30\x10\xa1\x0e\xa0\x0c\x30\x0a\xa0\x03\x02\x01\x02\xa1\x03\x01'
            b'\x01\xff',
            b'[{"b": {"c": [{"a": 2, "b": true}]}}]',
            b'\x01\x01\x81\x01\x01\x80\x01\x02\xff',
            b'\x30\x01\x80\x01\x02\x80',
            b'\x30\x18\x08\x14',
            b'<F><b><c><SEQUENCE><a>2</a><b><true /></b></SEQUENCE></c></b></'
            b'F>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'F', decoded, encoded)

        # H.
        decoded = [True]

        encoded_messages = [
            b'\x30\x03\x01\x01\xff',
            b'\x30\x03\x01\x01\xff',
            b'[true]',
            b'\x01\x01\xff',
            b'\x30',
            b'\x30',
            b'<H><true /></H>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'H', decoded, encoded)

        # I.
        decoded = [True]

        encoded_messages = [
            b'\x30\x03\x01\x01\xff',
            b'\x30\x03\x01\x01\xff',
            b'[true]',
            b'\x01\x01\xff',
            b'\xc0',
            b'\xc0',
            b'<I><true /></I>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'I', decoded, encoded)

        # S.
        decoded = {'a': True, 'b': 7}

        encoded_messages = [
            b'\x30\x08\xa0\x03\x01\x01\xff\x81\x01\x07',
            b'\x30\x08\xa0\x03\x01\x01\xff\x81\x01\x07',
            b'{"a": true, "b": 7}',
            b'\xff\x01\x07',
            b'\x80\x01\x07',
            b'\x80\x83\x80',
            b'<S><a><true /></a><b>7</b></S>'
        ]

        for spec, codec, encoded in zip(specs, CODECS, encoded_messages):
            self.encode_decode_codec(spec, codec, 'S', decoded, encoded)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'Parameterization': {'extensibility-implied': False,
                      'imports': {'Parameterization2': ['B2',
                                                        'D2',
                                                        'E2',
                                                        'F2']},
                      'object-classes': {},
                      'object-sets': {},
                      'tags': 'AUTOMATIC',
                      'types': {'A': {'members': [{'name': 'a', 'type': 'B'}],
                                      'parameters': ['B'],
                                      'type': 'SEQUENCE'},
                                'A-Boolean': {'actual-parameters': [{'type': 'BOOLEAN'}],
                                              'type': 'A'},
                                'A-Integer': {'actual-parameters': [{'type': 'INTEGER'}],
                                              'type': 'A'},
                                'A2': {'type': 'BOOLEAN'},
                                'B': {'members': [{'name': 'a', 'type': 'A'},
                                                  {'name': 'b',
                                                   'optional': True,
                                                   'type': 'B'}],
                                      'parameters': ['A', 'B'],
                                      'type': 'SEQUENCE'},
                                'B-BooleanInteger': {'actual-parameters': [{'type': 'BOOLEAN'},
                                                                           {'type': 'INTEGER'}],
                                                     'type': 'B'},
                                'C': {'members': [{'name': 'a', 'type': 'A'},
                                                  {'actual-parameters': [{'type': 'A'},
                                                                         {'type': 'INTEGER'}],
                                                   'name': 'b',
                                                   'type': 'B'}],
                                      'parameters': ['A'],
                                      'type': 'CHOICE'},
                                'D': {'members': [{'actual-parameters': [{'type': 'B-BooleanInteger'}],
                                                   'name': 'a',
                                                   'type': 'C'},
                                                  {'members': [{'actual-parameters': [{'actual-parameters': [{'type': 'NULL'},
                                                                                                             {'type': 'INTEGER'}],
                                                                                       'type': 'B'}],
                                                                'name': 'c',
                                                                'type': 'A'},
                                                               {'actual-parameters': [{'type': 'NULL'},
                                                                                      {'type': 'INTEGER'}],
                                                                'name': 'd',
                                                                'type': 'B'}],
                                                   'name': 'b',
                                                   'type': 'CHOICE'}],
                                      'type': 'SEQUENCE'},
                                'E': {'element': {'members': [{'name': 'a',
                                                               'type': 'A'},
                                                              {'members': [{'element': {'actual-parameters': [{'type': 'A'},
                                                                                                              {'type': 'BOOLEAN'}],
                                                                                        'type': 'B'},
                                                                            'name': 'c',
                                                                            'type': 'SEQUENCE '
                                                                                    'OF'}],
                                                               'name': 'b',
                                                               'type': 'SEQUENCE'}],
                                                  'type': 'CHOICE'},
                                      'parameters': ['A'],
                                      'size': [(0, 5)],
                                      'type': 'SEQUENCE OF'},
                                'F': {'actual-parameters': [{'type': 'INTEGER'}],
                                      'type': 'E'},
                                'G': {'element': {'type': 'BOOLEAN'},
                                      'parameters': ['a'],
                                      'size': [(0, 'a')],
                                      'type': 'SEQUENCE OF'},
                                'H': {'actual-parameters': [5], 'type': 'G'},
                                'I': {'actual-parameters': ['i'], 'type': 'G'},
                                'J': {'element': {'type': 'BOOLEAN'},
                                      'parameters': ['b'],
                                      'size': ['b'],
                                      'type': 'SEQUENCE OF'},
                                'K': {'actual-parameters': [3], 'type': 'J'},
                                'L': {'parameters': ['a', 'b'],
                                      'restricted-to': [('a', 'b')],
                                      'type': 'INTEGER'},
                                'M': {'actual-parameters': [3, 9], 'type': 'L'},
                                'N': {'parameters': ['a', 'B'],
                                      'restricted-to': [(-5, 'a')],
                                      'type': 'B'},
                                'O': {'actual-parameters': [-2,
                                                            {'type': 'INTEGER'}],
                                      'type': 'N'},
                                'P': {'members': [{'name': 'a', 'type': 'A'}],
                                      'parameters': ['A'],
                                      'type': 'SEQUENCE'},
                                'Q': {'actual-parameters': [{'type': 'A'}],
                                      'parameters': ['A'],
                                      'type': 'P'},
                                'R': {'actual-parameters': [{'type': 'INTEGER'}],
                                      'type': 'Q'},
                                'S': {'actual-parameters': [{'type': 'BOOLEAN'}],
                                      'type': 'B2'},
                                'T': {'actual-parameters': [{'type': 'OCTET '
                                                                     'STRING'}],
                                      'type': 'D2'},
                                'U': {'actual-parameters': [{'type': 'INTEGER'}],
                                      'type': 'E2'},
                                'V': {'actual-parameters': [{'type': 'INTEGER'}],
                                      'type': 'F2'}},
                      'values': {'i': {'type': 'INTEGER', 'value': 1}}},
 'Parameterization2': {'extensibility-implied': False,
                       'imports': {'Parameterization3': ['B3', 'D3']},
                       'object-classes': {},
                       'object-sets': {},
                       'tags': 'AUTOMATIC',
                       'types': {'A2': {'type': 'INTEGER'},
                                 'B2': {'members': [{'name': 'a', 'type': 'A'},
                                                    {'name': 'b',
                                                     'type': 'A2'}],
                                        'parameters': ['A'],
                                        'type': 'SEQUENCE'},
                                 'C2': {'type': 'INTEGER'},
                                 'D2': {'members': [{'name': 'a', 'type': 'A'},
                                                    {'name': 'b',
                                                     'type': 'C2'}],
                                        'parameters': ['A'],
                                        'type': 'SEQUENCE'},
                                 'E2': {'members': [{'name': 'a', 'type': 'A'},
                                                    {'actual-parameters': [{'type': 'A'}],
                                                     'name': 'b',
                                                     'type': 'B3'}],
                                        'parameters': ['A'],
                                        'type': 'SEQUENCE'},
                                 'F2': {'actual-parameters': [{'type': 'A'}],
                                        'parameters': ['A'],
                                        'type': 'D3'}},
                       'values': {}},
 'Parameterization3': {'extensibility-implied': False,
                       'imports': {},
                       'object-classes': {},
                       'object-sets': {},
                       'tags': 'AUTOMATIC',
                       'types': {'A3': {'type': 'INTEGER'},
                                 'B3': {'members': [{'name': 'a', 'type': 'A'},
                                                    {'name': 'b',
                                                     'type': 'A3'}],
                                        'parameters': ['A'],
                                        'type': 'SEQUENCE'},
                                 'C3': {'type': 'OCTET STRING'},
                                 'D3': {'parameters': ['A'], 'type': 'C3'}},
                       'values': {}}}


"""Basic Encoding Rules (BER) codec.

"""

import time
import math
import binascii
from copy import copy
import datetime

from ..errors import Error
from ..parser import EXTENSION_MARKER
from . import EncodeError
from . import DecodeError
from . import DecodeTagError
from . import DecodeContentsLengthError
from . import format_or
from . import compiler
from . import utc_time_to_datetime
from . import utc_time_from_datetime
from . import generalized_time_to_datetime
from . import generalized_time_from_datetime
from .compiler import enum_values_as_dict
from .compiler import clean_bit_string_value


class Class(object):
    UNIVERSAL        = 0x00
    APPLICATION      = 0x40
    CONTEXT_SPECIFIC = 0x80
    PRIVATE          = 0xc0


class Encoding(object):
    PRIMITIVE   = 0x00
    CONSTRUCTED = 0x20


class Tag(object):
    END_OF_CONTENTS   = 0x00
    BOOLEAN           = 0x01
    INTEGER           = 0x02
    BIT_STRING        = 0x03
    OCTET_STRING      = 0x04
    NULL              = 0x05
    OBJECT_IDENTIFIER = 0x06
    OBJECT_DESCRIPTOR = 0x07
    EXTERNAL          = 0x08
    REAL              = 0x09
    ENUMERATED        = 0x0a
    EMBEDDED_PDV      = 0x0b
    UTF8_STRING       = 0x0c
    RELATIVE_OID      = 0x0d
    SEQUENCE          = 0x10
    SET               = 0x11
    NUMERIC_STRING    = 0x12
    PRINTABLE_STRING  = 0x13
    T61_STRING        = 0x14
    VIDEOTEX_STRING   = 0x15
    IA5_STRING        = 0x16
    UTC_TIME          = 0x17
    GENERALIZED_TIME  = 0x18
    GRAPHIC_STRING    = 0x19
    VISIBLE_STRING    = 0x1a
    GENERAL_STRING    = 0x1b
    UNIVERSAL_STRING  = 0x1c
    CHARACTER_STRING  = 0x1d
    BMP_STRING        = 0x1e
    DATE              = 0x1f
    TIME_OF_DAY       = 0x20
    DATE_TIME         = 0x21


END_OF_CONTENTS_OCTETS = b'\x00\x00'


class DecodeChoiceError(Error):
    pass


def encode_length_definite(length):
    if length <= 127:
        encoded = bytearray([length])
    else:
        encoded = bytearray()

        while length > 0:
            encoded.append(length & 0xff)
            length >>= 8

        encoded.append(0x80 | len(encoded))
        encoded.reverse()

    return encoded


def decode_length_definite(encoded, offset):
    length = encoded[offset]
    offset += 1

    if length > 127:
        if length == 128:
            raise DecodeError(
                'Expected definite length at offset {}, but got indefinite.'.format(
                    offset - 1))

        number_of_bytes = (length & 0x7f)
        encoded_length = encoded[offset:number_of_bytes + offset]

        if len(encoded_length) != number_of_bytes:
            raise IndexError(
                'Expected {} length byte(s) at offset {}, but got {}.'.format(
                    number_of_bytes,
                    offset,
                    len(encoded_length)))

        length = int(binascii.hexlify(encoded_length), 16)
        offset += number_of_bytes

    if offset + length > len(encoded):
        raise DecodeContentsLengthError(length, offset, len(encoded))

    return length, offset


def decode_length_constructed(encoded, offset):
    length = encoded[offset]

    if length == 128:
        return None, offset + 1
    else:
        return decode_length_definite(encoded, offset)


def encode_signed_integer(data):
    encoded = bytearray()

    if data < 0:
        data *= -1
        data -= 1
        carry = not data

        while data > 0:
            encoded.append((data & 0xff) ^ 0xff)
            carry = (data & 0x80)
            data >>= 8

        if carry:
            encoded.append(0xff)
    elif data > 0:
        while data > 0:
            encoded.append(data & 0xff)
            data >>= 8

        if encoded[-1] & 0x80:
            encoded.append(0)
    else:
        encoded.append(0)

    encoded.reverse()

    return encoded


def decode_signed_integer(data):
    value = 0
    is_negative = (data[0] & 0x80)

    for byte in data:
        value <<= 8
        value += byte

    if is_negative:
        value -= (1 << (8 * len(data)))

    return value


def encode_tag(number, flags):
    if number < 31:
        tag = bytearray([flags | number])
    else:
        tag = bytearray([flags | 0x1f])
        encoded = bytearray()

        while number > 0:
            encoded.append(0x80 | (number & 0x7f))
            number >>= 7

        encoded[0] &= 0x7f
        encoded.reverse()
        tag.extend(encoded)

    return tag


def skip_tag(data, offset):
    byte = data[offset]
    offset += 1

    if byte & 0x1f == 0x1f:
        while data[offset] & 0x80:
            offset += 1

        offset += 1

    return offset


def read_tag(data, offset):
    return data[offset:skip_tag(data, offset)]


def skip_tag_length_contents(data, offset):
    offset = skip_tag(data, offset)

    return sum(decode_length_definite(data, offset))


def encode_real(data):
    if data == float('inf'):
        data = b'\x40'
    elif data == float('-inf'):
        data = b'\x41'
    elif math.isnan(data):
        data = b'\x42'
    elif data == 0.0:
        data = b''
    else:
        if data >= 0:
            negative_bit = 0
        else:
            negative_bit = 0x40
            data *= -1

        mantissa, exponent = math.frexp(abs(data))
        mantissa = int(mantissa * 2 ** 53)
        lowest_set_bit = compiler.lowest_set_bit(mantissa)
        mantissa >>= lowest_set_bit
        mantissa |= (0x80 << (8 * ((mantissa.bit_length() // 8) + 1)))
        mantissa = binascii.unhexlify(hex(mantissa)[4:].rstrip('L'))
        exponent = (52 - lowest_set_bit - exponent)

        if -129 < exponent < 128:
            exponent = [0x80 | negative_bit, ((0xff - exponent) & 0xff)]
        elif -32769 < exponent < 32768:
            exponent = ((0xffff - exponent) & 0xffff)
            exponent = [0x81 | negative_bit, (exponent >> 8), exponent & 0xff]
        else:
            raise NotImplementedError(
                'REAL exponent {} out of range.'.format(exponent))

        data = bytearray(exponent) + mantissa

    return data


def decode_real_binary(control, data):
    if control in [0x80, 0xc0]:
        exponent = data[1]

        if exponent & 0x80:
            exponent -= 0x100

        offset = 2
    elif control in [0x81, 0xc1]:
        exponent = ((data[1] << 8) | data[2])

        if exponent & 0x8000:
            exponent -= 0x10000

        offset = 3
    else:
        raise DecodeError(
            'Unsupported binary REAL control word 0x{:02x}.'.format(control))

    mantissa = int(binascii.hexlify(data[offset:]), 16)
    decoded = float(mantissa * 2 ** exponent)

    if control & 0x40:
        decoded *= -1

    return decoded


def decode_real_special(control):
    try:
        return {
            0x40: float('inf'),
            0x41: float('-inf'),
            0x42: float('nan'),
            0x43: -0.0
        }[control]
    except KeyError:
        raise DecodeError(
            'Unsupported special REAL control word 0x{:02x}.'.format(control))


def decode_real_decimal(data):
    return float(data[1:].replace(b',', b'.'))


def decode_real(data):
    if len(data) == 0:
        decoded = 0.0
    else:
        control = data[0]

        if control & 0x80:
            decoded = decode_real_binary(control, data)
        elif control & 0x40:
            decoded = decode_real_special(control)
        else:
            decoded = decode_real_decimal(data)

    return decoded


def encode_object_identifier(data):
    identifiers = [int(identifier) for identifier in data.split('.')]

    first_subidentifier = (40 * identifiers[0] + identifiers[1])
    encoded_subidentifiers = encode_object_identifier_subidentifier(
        first_subidentifier)

    for identifier in identifiers[2:]:
        encoded_subidentifiers += encode_object_identifier_subidentifier(
            identifier)

    return encoded_subidentifiers


def encode_object_identifier_subidentifier(subidentifier):
    encoded = [subidentifier & 0x7f]
    subidentifier >>= 7

    while subidentifier > 0:
        encoded.append(0x80 | (subidentifier & 0x7f))
        subidentifier >>= 7

    return encoded[::-1]


def decode_object_identifier(data, offset, end_offset):
    subidentifier, offset = decode_object_identifier_subidentifier(data,
                                                                   offset)
    decoded = [subidentifier // 40, subidentifier % 40]

    while offset < end_offset:
        subidentifier, offset = decode_object_identifier_subidentifier(data,
                                                                       offset)
        decoded.append(subidentifier)

    return '.'.join([str(v) for v in decoded])


def decode_object_identifier_subidentifier(data, offset):
    decoded = 0

    while data[offset] & 0x80:
        decoded += (data[offset] & 0x7f)
        decoded <<= 7
        offset += 1

    decoded += data[offset]

    return decoded, offset + 1


class Type(object):

    def __init__(self, name, type_name, number, flags=0):
        self.name = name
        self.type_name = type_name

        if number is None:
            self.tag = None
        else:
            self.tag = encode_tag(number, flags)

        self.optional = False
        self.default = None

    def set_tag(self, number, flags):
        self.tag = encode_tag(number, flags)

    def set_size_range(self, minimum, maximum, has_extension_marker):
        pass

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)

        if data[offset:end_offset] != self.tag:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

        return end_offset

    def set_default(self, value):
        self.default = value

    def get_default(self):
        return self.default

    def is_default(self, value):
        return value == self.default

    def has_default(self):
        return self.default is not None


class PrimitiveOrConstructedType(Type):

    def __init__(self, name, type_name, number, segment, flags=0):
        super(PrimitiveOrConstructedType, self).__init__(name,
                                                         type_name,
                                                         number,
                                                         flags)
        self.segment = segment
        self.constructed_tag = copy(self.tag)
        self.constructed_tag[0] |= Encoding.CONSTRUCTED

    def set_tag(self, number, flags):
        self.tag = encode_tag(number, flags)
        self.constructed_tag = copy(self.tag)
        self.constructed_tag[0] |= Encoding.CONSTRUCTED

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)
        tag = data[offset:end_offset]

        if tag == self.tag:
            return True, end_offset
        elif tag == self.constructed_tag:
            return False, end_offset
        else:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

    def decode(self, data, offset):
        is_primitive, offset = self.decode_tag(data, offset)

        if is_primitive:
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

            return self.decode_primitive_contents(data, offset, length), end_offset
        else:
            length, offset = decode_length_constructed(data, offset)
            segments = []

            if length is None:
                while data[offset:offset + 2] != END_OF_CONTENTS_OCTETS:
                    decoded, offset = self.segment.decode(data, offset)
                    segments.append(decoded)

                end_offset = offset + 2
            else:
                end_offset = offset + length

                while offset < end_offset:
                    decoded, offset = self.segment.decode(data, offset)
                    segments.append(decoded)

            return self.decode_constructed_segments(segments), end_offset

    def decode_primitive_contents(self, data, offset, length):
        raise NotImplementedError('To be implemented by subclasses.')

    def decode_constructed_segments(self, segments):
        raise NotImplementedError('To be implemented by subclasses.')


class StringType(PrimitiveOrConstructedType):

    TAG = None
    ENCODING = None

    def __init__(self, name):
        super(StringType, self).__init__(name,
                                         self.__class__.__name__,
                                         self.TAG,
                                         OctetString(name))

    def encode(self, data, encoded):
        data = data.encode(self.ENCODING)
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode(self.ENCODING)

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode(self.ENCODING)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.name)


class MembersType(Type):

    def __init__(self, name, tag_name, tag, root_members, additions):
        super(MembersType, self).__init__(name,
                                          tag_name,
                                          tag,
                                          Encoding.CONSTRUCTED)
        self.root_members = root_members
        self.additions = additions

    def set_tag(self, number, flags):
        super(MembersType, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_members = bytearray()

        for member in self.root_members:
            self.encode_member(member, data, encoded_members)

        if self.additions:
            self.encode_additions(data, encoded_members)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_members)))
        encoded.extend(encoded_members)

    def encode_additions(self, data, encoded_members):
        try:
            for addition in self.additions:
                encoded_addition = bytearray()

                if isinstance(addition, list):
                    for member in addition:
                        self.encode_member(member, data, encoded_addition)
                else:
                    self.encode_member(addition,
                                       data,
                                       encoded_addition)

                encoded_members.extend(encoded_addition)
        except EncodeError:
            pass

    def encode_member(self, member, data, encoded_members):
        name = member.name

        if name in data:
            value = data[name]

            try:
                if isinstance(member, AnyDefinedBy):
                    member.encode(value, encoded_members, data)
                elif not member.is_default(value):
                    member.encode(value, encoded_members)
            except EncodeError as e:
                e.location.append(member.name)
                raise
        elif member.optional:
            pass
        elif not member.has_default():
            raise EncodeError("{} member '{}' not found in {}.".format(
                self.__class__.__name__,
                name,
                data))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)

        if data[offset] == 0x80:
            # Indefinite length field.
            offset += 1
            end_offset = None
        else:
            # Definite length field
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

        values = {}

        for member in self.root_members:
            # End of indefinite length sequence may be reached at any
            # time, but DecodeError will occur (instead of usual
            # IndexError) and so further members will be skipped.
            offset = self.decode_member(member,
                                        data,
                                        values,
                                        offset,
                                        end_offset)

        if self.additions:
            offset = self.decode_additions(data,
                                           values,
                                           offset,
                                           end_offset)

        # Detect end of indefinite length constructed field.
        if end_offset is None:
            if data[offset:offset + 2] == END_OF_CONTENTS_OCTETS:
                end_offset = offset + 2
            else:
                raise DecodeError(
                    'Could not find end-of-contents tag for indefinite length '
                    'field.')

        return values, end_offset

    def decode_additions(self, data, values, offset, end_offset):
        try:
            for addition in self.additions:
                addition_values = {}

                if isinstance(addition, list):
                    for member in addition:
                        offset = self.decode_member(member,
                                                    data,
                                                    addition_values,
                                                    offset,
                                                    end_offset)
                else:
                    offset = self.decode_member(addition,
                                                data,
                                                addition_values,
                                                offset,
                                                end_offset)

                values.update(addition_values)
        except DecodeError:
            pass

        return offset

    def decode_member(self, member, data, values, offset, end_offset):
        try:
            # If reached end of indefinite length field, member.decode
            # will raise DecodeTagError, and end of field will be
            # handled in MembersType.decode() method.
            if end_offset is None or offset < end_offset:
                if isinstance(member, AnyDefinedBy):
                    value, offset = member.decode(data, offset, values)
                else:
                    value, offset = member.decode(data, offset)
            else:
                raise IndexError
        except (DecodeError, DecodeTagError, IndexError) as e:
            if member.optional:
                return offset

            if not member.has_default():
                if isinstance(e, IndexError):
                    e = DecodeError('out of data at offset {}'.format(offset))

                e.location.append(member.name)
                raise e

            value = member.get_default()

        values[member.name] = value

        return offset

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.root_members]))


class ArrayType(Type):

    def __init__(self, name, tag_name, tag, element_type):
        super(ArrayType, self).__init__(name,
                                        tag_name,
                                        tag,
                                        Encoding.CONSTRUCTED)
        self.element_type = element_type

    def set_tag(self, number, flags):
        super(ArrayType, self).set_tag(number,
                                       flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_elements = bytearray()

        for entry in data:
            self.element_type.encode(entry, encoded_elements)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_elements)))
        encoded.extend(encoded_elements)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        if data[offset] == 0x80:
            offset += 1
            length = None  # Indicates indefinite field.
        else:
            length, offset = decode_length_definite(data, offset)

        decoded = []
        start_offset = offset
        # Loop through data until length exceeded or end-of-contents tag reached.
        while True:
            if length is None:
                # Find end of indefinite sequence.
                if data[offset:offset + 2] == END_OF_CONTENTS_OCTETS:
                    offset += 2
                    break
            elif (offset - start_offset) >= length:
                # End of definite length sequence.
                break
            decoded_element, offset = self.element_type.decode(data, offset)
            decoded.append(decoded_element)

        return decoded, offset

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name,
                                      'BOOLEAN',
                                      Tag.BOOLEAN)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(1)
        encoded.append(0xff * data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, contents_offset = decode_length_definite(data, offset)

        if length != 1:
            raise DecodeError(
                'Expected BOOLEAN contents length 1 at offset {}, but '
                'got {}.'.format(offset,
                                 length))

        return bool(data[contents_offset]), contents_offset + length

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name,
                                      'INTEGER',
                                      Tag.INTEGER)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        value = encode_signed_integer(data)
        encoded.extend(encode_length_definite(len(value)))
        encoded.extend(value)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return decode_signed_integer(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL', Tag.REAL)

    def encode(self, data, encoded):
        data = encode_real(data)
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = decode_real(data[offset:end_offset])

        return decoded, end_offset

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL', Tag.NULL)

    def is_default(self, value):
        return False

    def encode(self, _, encoded):
        encoded.extend(self.tag)
        encoded.append(0)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)

        return None, offset + 1

    def __repr__(self):
        return 'Null({})'.format(self.name)


class BitString(PrimitiveOrConstructedType):

    def __init__(self, name, has_named_bits):
        super(BitString, self).__init__(name,
                                        'BIT STRING',
                                        Tag.BIT_STRING,
                                        self)
        self.has_named_bits = has_named_bits

    def is_default(self, value):
        if self.default is None:
            return False

        clean_value = clean_bit_string_value(value,
                                             self.has_named_bits)
        clean_default = clean_bit_string_value(self.default,
                                               self.has_named_bits)

        return clean_value == clean_default

    def encode(self, data, encoded):
        number_of_bytes, number_of_rest_bits = divmod(data[1], 8)
        data = bytearray(data[0])

        if number_of_rest_bits == 0:
            data = data[:number_of_bytes]
            number_of_unused_bits = 0
        else:
            last_byte = data[number_of_bytes]
            last_byte &= ((0xff >> number_of_rest_bits) ^ 0xff)
            data = data[:number_of_bytes]
            data.append(last_byte)
            number_of_unused_bits = (8 - number_of_rest_bits)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data) + 1))
        encoded.append(number_of_unused_bits)
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        length -= 1
        number_of_bits = 8 * length - data[offset]
        offset += 1

        return (data[offset:offset + length], number_of_bits)

    def decode_constructed_segments(self, segments):
        decoded = bytearray()
        number_of_bits = 0

        for data, length in segments:
            decoded.extend(data)
            number_of_bits += length

        return (bytes(decoded), number_of_bits)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(OctetString, self).__init__(name,
                                          'OCTET STRING',
                                          Tag.OCTET_STRING,
                                          self)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        return bytes(data[offset:offset + length])

    def decode_constructed_segments(self, segments):
        return bytes().join(segments)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name,
                                               'OBJECT IDENTIFIER',
                                               Tag.OBJECT_IDENTIFIER)

    def encode(self, data, encoded):
        encoded_subidentifiers = encode_object_identifier(data)
        encoded.extend(self.tag)
        encoded.append(len(encoded_subidentifiers))
        encoded.extend(encoded_subidentifiers)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = decode_object_identifier(data, offset, end_offset)

        return decoded, end_offset

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values, numeric):
        super(Enumerated, self).__init__(name,
                                         'ENUMERATED',
                                         Tag.ENUMERATED)
        if numeric:
            self.value_to_data = {k: k for k in enum_values_as_dict(values)}
            self.data_to_value = self.value_to_data
        else:
            self.value_to_data = enum_values_as_dict(values)
            self.data_to_value = {v: k for k, v in self.value_to_data.items()}

        self.has_extension_marker = (EXTENSION_MARKER in values)

    def format_names(self):
        return format_or(sorted(list(self.value_to_data.values())))

    def format_values(self):
        return format_or(sorted(list(self.value_to_data)))

    def encode(self, data, encoded):
        try:
            value = self.data_to_value[data]
        except KeyError:
            raise EncodeError(
                "Expected enumeration value {}, but got '{}'.".format(
                    self.format_names(),
                    data))

        encoded.extend(self.tag)
        value = encode_signed_integer(value)
        encoded.extend(encode_length_definite(len(value)))
        encoded.extend(value)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        value = decode_signed_integer(data[offset:end_offset])

        if value in self.value_to_data:
            return self.value_to_data[value], end_offset
        elif self.has_extension_marker:
            return None, end_offset
        else:
            raise DecodeError(
                'Expected enumeration value {}, but got {}.'.format(
                    self.format_values(),
                    value))

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class Sequence(MembersType):

    def __init__(self, name, root_members, additions):
        super(Sequence, self).__init__(name,
                                       'SEQUENCE',
                                       Tag.SEQUENCE,
                                       root_members,
                                       additions)


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         'SEQUENCE OF',
                                         Tag.SEQUENCE,
                                         element_type)


class Set(MembersType):

    def __init__(self, name, root_members, additions):
        super(Set, self).__init__(name,
                                  'SET',
                                  Tag.SET,
                                  root_members,
                                  additions)


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    'SET OF',
                                    Tag.SET,
                                    element_type)


class Choice(Type):

    def __init__(self, name, root_members, additions):
        super(Choice, self).__init__(name, 'CHOICE', None)
        members = root_members

        if additions is not None:
            for addition in additions:
                if isinstance(addition, list):
                    members += addition
                else:
                    members.append(addition)

            self.has_extension_marker = True
        else:
            self.has_extension_marker = False

        self.members = members
        self.name_to_member = {member.name: member for member in self.members}
        self.tag_to_member = {}
        self.add_tags(self.members)

    def add_tags(self, members):
        for member in members:
            tags = self.get_member_tags(member)

            for tag in tags:
                self.tag_to_member[tag] = member

    def get_member_tags(self, member):
        tags = []

        if isinstance(member, Choice):
            tags = self.get_choice_tags(member)
        elif isinstance(member, Recursive):
            if member.inner is None:
                member.choice_parents.append(self)
            else:
                tags = self.get_member_tags(member.inner)
        else:
            tags.append(bytes(member.tag))

            if hasattr(member, 'constructed_tag'):
                tags.append(bytes(member.constructed_tag))

        return tags

    def get_choice_tags(self, choice):
        tags = []

        for member in choice.members:
            tags.extend(self.get_member_tags(member))

        return tags

    def format_tag(self, tag):
        return binascii.hexlify(tag).decode('ascii')

    def format_tags(self):
        return format_or(sorted([self.format_tag(tag)
                                 for tag in self.tag_to_member]))

    def format_names(self):
        return format_or(sorted([member.name for member in self.members]))

    def encode(self, data, encoded):
        try:
            member = self.name_to_member[data[0]]
        except KeyError:
            raise EncodeError(
                "Expected choice {}, but got '{}'.".format(
                    self.format_names(),
                    data[0]))

        try:
            member.encode(data[1], encoded)
        except EncodeError as e:
            e.location.append(member.name)
            raise

    def decode(self, data, offset):
        tag = bytes(read_tag(data, offset))

        if tag in self.tag_to_member:
            member = self.tag_to_member[tag]
        elif self.has_extension_marker:
            offset = skip_tag_length_contents(data, offset)

            return (None, None), offset
        else:
            raise DecodeError(
                "Expected choice member tag {}, but got '{}'.".format(
                    self.format_tags(),
                    self.format_tag(tag)))

        decoded, offset = member.decode(data, offset)

        return (member.name, decoded), offset

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class UTF8String(StringType):

    TAG = Tag.UTF8_STRING
    ENCODING = 'utf-8'


class NumericString(StringType):

    TAG = Tag.NUMERIC_STRING
    ENCODING = 'ascii'


class PrintableString(StringType):

    TAG = Tag.PRINTABLE_STRING
    ENCODING = 'ascii'


class IA5String(StringType):

    TAG = Tag.IA5_STRING
    ENCODING = 'ascii'


class VisibleString(StringType):

    TAG = Tag.VISIBLE_STRING
    ENCODING = 'ascii'


class GeneralString(StringType):

    TAG = Tag.GENERAL_STRING
    ENCODING = 'latin-1'


class BMPString(StringType):

    TAG = Tag.BMP_STRING
    ENCODING = 'utf-16-be'


class GraphicString(StringType):

    TAG = Tag.GRAPHIC_STRING
    ENCODING = 'latin-1'


class UniversalString(StringType):

    TAG = Tag.UNIVERSAL_STRING
    ENCODING = 'utf-32-be'


class TeletexString(StringType):

    TAG = Tag.T61_STRING
    ENCODING = 'iso-8859-1'


class ObjectDescriptor(GraphicString):

    TAG = Tag.OBJECT_DESCRIPTOR


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name,
                                      'UTCTime',
                                      Tag.UTC_TIME)

    def encode(self, data, encoded):
        data = utc_time_from_datetime(data).encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')

        return utc_time_to_datetime(decoded), end_offset

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name,
                                              'GeneralizedTime',
                                              Tag.GENERALIZED_TIME)

    def encode(self, data, encoded):
        data = generalized_time_from_datetime(data).encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')

        return generalized_time_to_datetime(decoded), end_offset

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class Date(Type):

    def __init__(self, name):
        super(Date, self).__init__(name, 'DATE', Tag.DATE)

    def encode(self, data, encoded):
        data = str(data).replace('-', '').encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')
        decoded = datetime.date(*time.strptime(decoded, '%Y%m%d')[:3])

        return decoded, end_offset

    def __repr__(self):
        return 'Date({})'.format(self.name)


class TimeOfDay(Type):

    def __init__(self, name):
        super(TimeOfDay, self).__init__(name,
                                        'TIME-OF-DAY',
                                        Tag.TIME_OF_DAY)

    def encode(self, data, encoded):
        data = str(data).replace(':', '').encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')
        decoded = datetime.time(*time.strptime(decoded, '%H%M%S')[3:6])

        return decoded, end_offset

    def __repr__(self):
        return 'TimeOfDay({})'.format(self.name)


class DateTime(Type):

    def __init__(self, name):
        super(DateTime, self).__init__(name,
                                       'DATE-TIME',
                                       Tag.DATE_TIME)

    def encode(self, data, encoded):
        data = '{:04d}{:02d}{:02d}{:02d}{:02d}{:02d}'.format(*data.timetuple())
        data = data.encode('ascii')
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = data[offset:end_offset].decode('ascii')
        decoded = datetime.datetime(*time.strptime(decoded, '%Y%m%d%H%M%S')[:6])

        return decoded, end_offset

    def __repr__(self):
        return 'DateTime({})'.format(self.name)


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY', None)

    def encode(self, data, encoded):
        encoded.extend(data)

    def decode(self, data, offset):
        start = offset
        offset = skip_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[start:end_offset], end_offset

    def __repr__(self):
        return 'Any({})'.format(self.name)


class AnyDefinedBy(Type):

    def __init__(self, name, type_member, choices):
        super(AnyDefinedBy, self).__init__(name,
                                           'ANY DEFINED BY',
                                           None,
                                           None)
        self.type_member = type_member
        self.choices = choices

    def encode(self, data, encoded, values):
        if self.choices:
            try:
                self.choices[values[self.type_member]].encode(data, encoded)
            except KeyError:
                raise EncodeError('Bad AnyDefinedBy choice {}.'.format(
                    values[self.type_member]))
        else:
            encoded.extend(data)

    def decode(self, data, offset, values):
        if self.choices:
            try:
                return self.choices[values[self.type_member]].decode(data,
                                                                     offset)
            except KeyError:
                raise DecodeError('Bad AnyDefinedBy choice {}.'.format(
                    values[self.type_member]))
        else:
            start = offset
            offset = skip_tag(data, offset)
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

            return data[start:end_offset], end_offset

    def __repr__(self):
        return 'AnyDefinedBy({})'.format(self.name)


class ExplicitTag(Type):

    def __init__(self, name, inner):
        super(ExplicitTag, self).__init__(name, 'ExplicitTag', None)
        self.inner = inner

    def set_default(self, value):
        self.inner.set_default(value)

    def get_default(self):
        return self.inner.get_default()

    def is_default(self, value):
        return self.inner.is_default(value)

    def has_default(self):
        return self.inner.has_default()

    def set_tag(self, number, flags):
        super(ExplicitTag, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_inner = bytearray()
        self.inner.encode(data, encoded_inner)
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_inner)))
        encoded.extend(encoded_inner)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        indefinite = False

        if data[offset] == 0x80:
            # Indefinite length field
            offset += 1
            indefinite = True
        else:
            # Definite length field
            length, offset = decode_length_definite(data, offset)

        values, end_offset = self.inner.decode(data, offset)

        if indefinite:
            if data[end_offset:end_offset + 2] != END_OF_CONTENTS_OCTETS:
                raise DecodeError(
                    'Expected end-of-contents tag at offset: {}.'.format(
                        end_offset))

            end_offset += 2

        return values, end_offset

    def __repr__(self):
        return 'ExplicitTag({})'.format(self.inner)


class Recursive(Type, compiler.Recursive):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE', None)
        self.type_name = type_name
        self.module_name = module_name
        self.tag_number = None
        self.tag_flags = None
        self.inner = None
        self.choice_parents = []

    def set_tag(self, number, flags):
        self.tag_number = number
        self.tag_flags = flags

    def set_inner_type(self, inner):
        self.inner = copy(inner)

        if self.tag_number is not None:
            self.inner.set_tag(self.tag_number, self.tag_flags)

        for choice_parent in self.choice_parents:
            choice_parent.add_tags([self])

    def encode(self, data, encoded):
        self.inner.encode(data, encoded)

    def decode(self, data, offset):
        return self.inner.decode(data, offset)

    def __repr__(self):
        return 'Recursive({})'.format(self.type_name)


class CompiledType(compiler.CompiledType):

    def encode(self, data):
        encoded = bytearray()
        self._type.encode(data, encoded)

        return encoded

    def decode(self, data):
        return self._type.decode(bytearray(data), 0)[0]

    def decode_with_length(self, data):
        return self._type.decode(bytearray(data), 0)


def get_tag_no_encoding(member):
    value = (member.tag[0] & ~Encoding.CONSTRUCTED)

    return bytearray([value]) + member.tag[1:]


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        compiled_type = self.compile_type(type_name,
                                          type_descriptor,
                                          module_name)

        return CompiledType(compiled_type)

    def compile_implicit_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name,
                                      sort_by_tag=True))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(
                name,
                *self.compile_members(type_descriptor['members'],
                                      module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name,
                                  self.get_enum_values(type_descriptor,
                                                       module_name),
                                  self._numeric_enums)
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'DATE':
            compiled = Date(name)
        elif type_name == 'TIME-OF-DAY':
            compiled = TimeOfDay(name)
        elif type_name == 'DATE-TIME':
            compiled = DateTime(name)
        elif type_name == 'BIT STRING':
            has_named_bits = ('named-bits' in type_descriptor)
            compiled = BitString(name, has_named_bits)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            choices = {}

            for key, value in type_descriptor['choices'].items():
                choices[key] = self.compile_type(key,
                                                 value,
                                                 module_name)

            compiled = AnyDefinedBy(name,
                                    type_descriptor['value'],
                                    choices)
        elif type_name == 'NULL':
            compiled = Null(name)
        elif type_name == 'EXTERNAL':
            compiled = Sequence(
                name,
                *self.compile_members(self.external_type_descriptor()['members'],
                                      module_name))
            compiled.set_tag(Tag.EXTERNAL, 0)
        elif type_name == 'ObjectDescriptor':
            compiled = ObjectDescriptor(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
                self.recursive_types.append(compiled)
            else:
                compiled = self.compile_user_type(name,
                                                  type_name,
                                                  module_name)

        return compiled

    def compile_type(self, name, type_descriptor, module_name):
        module_name = self.get_module_name(type_descriptor, module_name)
        compiled = self.compile_implicit_type(name,
                                              type_descriptor,
                                              module_name)

        if self.is_explicit_tag(type_descriptor):
            compiled = ExplicitTag(name, compiled)

        # Set any given tag.
        if 'tag' in type_descriptor:
            compiled = self.copy(compiled)
            tag = type_descriptor['tag']
            class_ = tag.get('class', None)

            if class_ == 'APPLICATION':
                flags = Class.APPLICATION
            elif class_ == 'PRIVATE':
                flags = Class.PRIVATE
            elif class_ == 'UNIVERSAL':
                flags = 0
            else:
                flags = Class.CONTEXT_SPECIFIC

            compiled.set_tag(tag['number'], flags)

        return compiled

    def compile_members(self,
                        members,
                        module_name,
                        sort_by_tag=False):
        compiled_members = []
        in_extension = False
        additions = None

        for member in members:
            if member == EXTENSION_MARKER:
                in_extension = not in_extension

                if in_extension:
                    additions = []
            elif in_extension:
                self.compile_extension_member(member,
                                              module_name,
                                              additions)
            else:
                self.compile_root_member(member,
                                         module_name,
                                         compiled_members)

        if sort_by_tag:
            compiled_members = sorted(compiled_members, key=get_tag_no_encoding)

        return compiled_members, additions

    def compile_extension_member(self,
                                 member,
                                 module_name,
                                 additions):
        if isinstance(member, list):
            compiled_group = []

            for memb in member:
                compiled_member = self.compile_member(memb,
                                                      module_name)
                compiled_group.append(compiled_member)

            additions.append(compiled_group)
        else:
            compiled_member = self.compile_member(member,
                                                  module_name)
            additions.append(compiled_member)


def compile_dict(specification, numeric_enums=False):
    return Compiler(specification, numeric_enums).process()


def decode_length(data):
    try:
        return skip_tag_length_contents(bytearray(data), 0)
    except DecodeContentsLengthError as e:
        return (e.length + e.offset)
    except IndexError:
        return None
